#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import base64
import binascii
import hashlib
import logging
import os
import subprocess
import sys
import tempfile
from base64 import b64decode
from typing import Optional, Tuple

from construct import (AlignedStruct, Bytes, GreedyRange, Int32ul, Padding,
                       Struct, abs_, this)

from . import ESPCoreDumpLoaderError
from .elf import (TASK_STATUS_CORRECT, TASK_STATUS_TCB_CORRUPTED, ElfFile,
                  ElfSegment, ESPCoreDumpElfFile, EspTaskStatus, NoteSection)
from .riscv import (Esp32C2Methods, Esp32C3Methods, Esp32C5Methods,
                    Esp32C6Methods, Esp32C61Methods, Esp32H2Methods,
                    Esp32P4Methods)
from .xtensa import Esp32Methods, Esp32S2Methods, Esp32S3Methods

IDF_PATH = os.getenv('IDF_PATH', '')
PARTTOOL_PY = os.path.join(IDF_PATH, 'components', 'partition_table', 'parttool.py')

# Following structs are based on source code
# components/espcoredump/include_core_dump/esp_core_dump_priv.h

EspCoreDumpV1Header = Struct(
    'tot_len' / Int32ul,
    'ver' / Int32ul,
    'task_num' / Int32ul,
    'tcbsz' / Int32ul,
)

EspCoreDumpV2Header = Struct(
    'tot_len' / Int32ul,
    'ver' / Int32ul,
    'task_num' / Int32ul,
    'tcbsz' / Int32ul,
    'segs_num' / Int32ul,
)

EspCoreDumpV2_1_Header = Struct(
    'tot_len' / Int32ul,
    'ver' / Int32ul,
    'task_num' / Int32ul,
    'tcbsz' / Int32ul,
    'segs_num' / Int32ul,
    'chip_rev' / Int32ul,
)

CRC = Int32ul
SHA256 = Bytes(32)

TaskHeader = Struct(
    'tcb_addr' / Int32ul,
    'stack_top' / Int32ul,
    'stack_end' / Int32ul,
)

MemSegmentHeader = Struct(
    'mem_start' / Int32ul,
    'mem_sz' / Int32ul,
    'data' / Bytes(this.mem_sz),
)


def get_core_file_format(core_file: str) -> str:
    """Get format of core_file based on the header"""
    with open(core_file, 'rb') as f:
        coredump_bytes = f.read(16)

        # Check if this is an ELF file without the core dump header (core_dump_header_t)
        if coredump_bytes.startswith(b'\x7fELF'):
            return 'elf'

        # Check if this is a core dump with a core_dump_header_t header
        core_version = EspCoreDumpVersion(int.from_bytes(coredump_bytes[4:7], 'little'))
        if core_version.dump_ver in EspCoreDumpLoader.CORE_VERSIONS:
            return 'raw'

    # Neither of theses headers matched, so this might be a base64 encoded core dump;
    # however in case it's just some unknown binary, ignore decoding errors.
    with open(core_file, 'r', encoding='utf-8', errors='ignore') as c:
        coredump_str = c.read()
        try:
            b64decode(coredump_str)
        except Exception:
            raise SystemExit(
                'The format of the provided core-file is not recognized. '
                'Please ensure that the core-format matches one of the following: ELF (“elf”), '
                'raw (raw) or base64-encoded (b64) binary'
            )
        return 'b64'


class EspCoreDumpVersion(object):
    """Core dump version class, it contains all version-dependent params
    """
    # Chip IDs should be in sync with components/esp_hw_support/include/esp_chip_info.h
    ESP32 = 0
    ESP32S2 = 2
    ESP32S3 = 9
    XTENSA_CHIPS = [ESP32, ESP32S2, ESP32S3]

    ESP32C3 = 5
    ESP32C2 = 12
    ESP32C6 = 13
    ESP32H2 = 16
    ESP32P4 = 18
    ESP32C61 = 20
    ESP32C5 = 23
    RISCV_CHIPS = [ESP32C3, ESP32C2, ESP32H2, ESP32C6, ESP32P4, ESP32C5, ESP32C61]

    COREDUMP_SUPPORTED_TARGETS = XTENSA_CHIPS + RISCV_CHIPS

    def __init__(self, version=None):  # type: (int) -> None
        """Constructor for core dump version
        """
        super(EspCoreDumpVersion, self).__init__()
        if version is None:
            self.version = 0
        else:
            self.set_version(version)

    @staticmethod
    def make_dump_ver(major, minor):  # type: (int, int) -> int
        return ((major & 0xFF) << 8) | ((minor & 0xFF) << 0)

    def set_version(self, version):  # type: (int) -> None
        self.version = version

    @property
    def chip_ver(self):  # type: () -> int
        return (self.version & 0xFFFF0000) >> 16

    @property
    def dump_ver(self):  # type: () -> int
        return self.version & 0x0000FFFF

    @property
    def major(self):  # type: () -> int
        return (self.version & 0x0000FF00) >> 8

    @property
    def minor(self):  # type: () -> int
        return self.version & 0x000000FF


class EspCoreDumpLoader(EspCoreDumpVersion):
    # "legacy" stands for core dumps v0.1 (before IDF v4.1)
    BIN_V1 = EspCoreDumpVersion.make_dump_ver(0, 1)
    BIN_V2 = EspCoreDumpVersion.make_dump_ver(0, 2)
    BIN_V2_1 = EspCoreDumpVersion.make_dump_ver(0, 3)
    ELF_CRC32_V2 = EspCoreDumpVersion.make_dump_ver(1, 0)
    ELF_CRC32_V2_1 = EspCoreDumpVersion.make_dump_ver(1, 2)
    ELF_SHA256_V2 = EspCoreDumpVersion.make_dump_ver(1, 1)
    ELF_SHA256_V2_1 = EspCoreDumpVersion.make_dump_ver(1, 3)
    CORE_VERSIONS = [BIN_V1, BIN_V2, BIN_V2_1, ELF_CRC32_V2, ELF_CRC32_V2_1, ELF_SHA256_V2, ELF_SHA256_V2_1]

    def __init__(self):  # type: () -> None
        super(EspCoreDumpLoader, self).__init__()
        self.core_src_file = None  # type: Optional[str]
        self.core_src_struct = None
        self.core_src = None

        self.core_elf_file: str = None  # type: ignore

        self.header = None
        self.header_struct = EspCoreDumpV1Header
        self.checksum_struct = CRC

        # target classes will be assigned in ``_reload_coredump``
        self.target_methods = Esp32Methods()

        self.temp_files = []  # type: list[str]

    def _create_temp_file(self):  # type: () -> str
        t = tempfile.NamedTemporaryFile('wb', delete=False)
        # Here we close this at first to make sure the read/write is wrapped in context manager
        # Otherwise the result will be wrong if you read while open in another session
        t.close()
        self.temp_files.append(t.name)
        return t.name

    def _load_core_src(self):  # type: () -> str
        """
        Write core elf into ``self.core_src``,
        Return the target str by reading core elf
        """
        with open(self.core_src_file, 'rb') as fr:  # type: ignore
            coredump_bytes = fr.read()

        _header = EspCoreDumpV1Header.parse(coredump_bytes)  # first we use V1 format to get version
        self.set_version(_header.ver)
        if self.dump_ver == self.ELF_CRC32_V2:
            self.checksum_struct = CRC
            self.header_struct = EspCoreDumpV2Header
        elif self.dump_ver == self.ELF_CRC32_V2_1:
            self.checksum_struct = CRC
            self.header_struct = EspCoreDumpV2_1_Header
        elif self.dump_ver == self.ELF_SHA256_V2:
            self.checksum_struct = SHA256
            self.header_struct = EspCoreDumpV2Header
        elif self.dump_ver == self.ELF_SHA256_V2_1:
            self.checksum_struct = SHA256
            self.header_struct = EspCoreDumpV2_1_Header
        elif self.dump_ver == self.BIN_V1:
            self.checksum_struct = CRC
            self.header_struct = EspCoreDumpV1Header
        elif self.dump_ver == self.BIN_V2:
            self.checksum_struct = CRC
            self.header_struct = EspCoreDumpV2Header
        elif self.dump_ver == self.BIN_V2_1:
            self.checksum_struct = CRC
            self.header_struct = EspCoreDumpV2_1_Header
        else:
            raise ESPCoreDumpLoaderError('Core dump version "0x%x" is not supported!' % self.dump_ver)

        self.header = self.header_struct.parse(coredump_bytes)

        self.core_src_struct = Struct(
            'header' / self.header_struct,
            'data' / Bytes(this.header.tot_len - self.header_struct.sizeof() - self.checksum_struct.sizeof()),
            'checksum' / self.checksum_struct,
        )
        self.core_src = self.core_src_struct.parse(coredump_bytes)  # type: ignore

        if self.header and self.header.get('chip_rev') is not None:
            self.chip_rev = self.header.chip_rev  # type: ignore
        else:
            self.chip_rev = None  # type: ignore

        if self.chip_ver in self.COREDUMP_SUPPORTED_TARGETS:
            if self.chip_ver == self.ESP32:
                self.target_methods = Esp32Methods()  # type: ignore
            elif self.chip_ver == self.ESP32S2:
                self.target_methods = Esp32S2Methods()  # type: ignore
            elif self.chip_ver == self.ESP32C3:
                self.target_methods = Esp32C3Methods()  # type: ignore
            elif self.chip_ver == self.ESP32S3:
                self.target_methods = Esp32S3Methods()  # type: ignore
            elif self.chip_ver == self.ESP32C2:
                self.target_methods = Esp32C2Methods()  # type: ignore
            elif self.chip_ver == self.ESP32H2:
                self.target_methods = Esp32H2Methods()  # type: ignore
            elif self.chip_ver == self.ESP32C6:
                self.target_methods = Esp32C6Methods()  # type: ignore
            elif self.chip_ver == self.ESP32P4:
                self.target_methods = Esp32P4Methods()  # type: ignore
            elif self.chip_ver == self.ESP32C5:
                self.target_methods = Esp32C5Methods()  # type: ignore
            elif self.chip_ver == self.ESP32C61:
                self.target_methods = Esp32C61Methods()  # type: ignore
            else:
                raise NotImplementedError
        else:
            raise ESPCoreDumpLoaderError('Core dump chip "0x%x" is not supported!' % self.chip_ver)

        return self.target_methods.TARGET  # type: ignore

    def _validate_dump_file(self):  # type: () -> None
        if self.chip_ver not in self.COREDUMP_SUPPORTED_TARGETS:
            raise ESPCoreDumpLoaderError('Invalid core dump chip version: "{}", should be <= "0x{:X}"'
                                         .format(self.chip_ver, self.ESP32S2))
        if self.checksum_struct == CRC:
            self._crc_validate()
        elif self.checksum_struct == SHA256:
            self._sha256_validate()

    def _crc_validate(self):  # type: () -> None
        if self.dump_ver in [self.BIN_V2_1,
                             self.ELF_CRC32_V2_1]:
            data_crc = binascii.crc32(
                EspCoreDumpV2_1_Header.build(self.core_src.header) + self.core_src.data) & 0xffffffff  # type: ignore
        else:
            data_crc = binascii.crc32(
                EspCoreDumpV2Header.build(self.core_src.header) + self.core_src.data) & 0xffffffff  # type: ignore
        if data_crc != self.core_src.checksum:  # type: ignore
            raise ESPCoreDumpLoaderError(
                'Invalid core dump CRC %x, should be %x' % (data_crc, self.core_src.checksum))  # type: ignore

    def _sha256_validate(self):  # type: () -> None
        if self.dump_ver in [self.ELF_SHA256_V2_1]:
            data_sha256 = hashlib.sha256(
                EspCoreDumpV2_1_Header.build(self.core_src.header) + self.core_src.data)  # type: ignore
        else:
            data_sha256 = hashlib.sha256(
                EspCoreDumpV2Header.build(self.core_src.header) + self.core_src.data)  # type: ignore

        data_sha256_str = data_sha256.hexdigest()
        sha256_str = binascii.hexlify(self.core_src.checksum).decode('ascii')  # type: ignore
        if data_sha256_str != sha256_str:
            raise ESPCoreDumpLoaderError('Invalid core dump SHA256 "{}", should be "{}"'
                                         .format(data_sha256_str, sha256_str))

    def create_corefile(self, exe_name=None, e_machine=ESPCoreDumpElfFile.EM_XTENSA):
        # type: (Optional[str], Optional[int]) -> None
        """
        Creates core dump ELF file
        """
        self._validate_dump_file()
        self.core_elf_file = self._create_temp_file()

        if self.dump_ver in [self.ELF_CRC32_V2,
                             self.ELF_CRC32_V2_1,
                             self.ELF_SHA256_V2,
                             self.ELF_SHA256_V2_1]:
            self._extract_elf_corefile(exe_name, e_machine)
        elif self.dump_ver in [self.BIN_V1,
                               self.BIN_V2,
                               self.BIN_V2_1]:
            self._extract_bin_corefile(e_machine)
        else:
            raise NotImplementedError

    def _extract_elf_corefile(self, exe_name=None, e_machine=ESPCoreDumpElfFile.EM_XTENSA):
        # type: (str, Optional[int]) -> None
        """
        Reads the ELF formatted core dump image and parse it
        """
        with open(self.core_elf_file, 'wb') as fw:  # type: ignore
            fw.write(self.core_src.data)  # type: ignore

        core_elf = ESPCoreDumpElfFile(self.core_elf_file, e_machine=e_machine)  # type: ignore

        if self.chip_rev is not None:  # type: ignore
            chip_rev_note = b''
            chip_rev_note += self._build_note_section('ESP_CHIP_REV',
                                                      ESPCoreDumpElfFile.PT_ESP_INFO,
                                                      Int32ul.build(self.chip_rev))  # type: ignore
            try:
                core_elf.add_segment(0, chip_rev_note, ElfFile.PT_NOTE, 0)
            except ESPCoreDumpLoaderError as e:
                logging.warning('Skip core dump info NOTES segment {:d} bytes @ 0x{:x}. (Reason: {})'
                                .format(len(chip_rev_note), 0, e))
            core_elf.dump(self.core_elf_file)

        # Read note segments from core file which are belong to tasks (TCB or stack)
        for seg in core_elf.note_segments:
            for note_sec in seg.note_secs:
                # Check for version info note
                if note_sec.name == b'ESP_CORE_DUMP_INFO' \
                        and note_sec.type == ESPCoreDumpElfFile.PT_ESP_INFO \
                        and exe_name:
                    exe_elf = ElfFile(exe_name)
                    app_sha256 = binascii.hexlify(exe_elf.sha256)
                    coredump_sha256_struct = Struct(
                        'ver' / Int32ul,
                        'sha256' / Bytes(64)  # SHA256 as hex string
                    )
                    coredump_sha256 = coredump_sha256_struct.parse(note_sec.desc[:coredump_sha256_struct.sizeof()])

                    logging.debug('App SHA256: {!r}'.format(app_sha256))
                    logging.debug('Core dump SHA256: {!r}'.format(coredump_sha256))

                    # Actual coredump SHA may be shorter than a full SHA256 hash
                    # with NUL byte padding, according to the app's APP_RETRIEVE_LEN_ELF_SHA
                    # length
                    core_sha_trimmed = coredump_sha256.sha256.rstrip(b'\x00').decode()
                    app_sha_trimmed = app_sha256[:len(core_sha_trimmed)].decode()

                    if core_sha_trimmed != app_sha_trimmed:
                        raise ESPCoreDumpLoaderError(
                            'Invalid application image for coredump: coredump SHA256({}) != app SHA256({}).'
                            .format(core_sha_trimmed, app_sha_trimmed))
                    if coredump_sha256.ver != self.version:
                        raise ESPCoreDumpLoaderError(
                            'Invalid application image for coredump: coredump SHA256 version({}) != app SHA256 version({}).'
                            .format(coredump_sha256.ver, self.version))

    @staticmethod
    def _get_aligned_size(size, align_with=4):  # type: (int, int) -> int
        if size % align_with:
            return align_with * (size // align_with + 1)
        return size

    @staticmethod
    def _build_note_section(name, sec_type, desc):  # type: (str, int, str) -> bytes
        b_name = bytearray(name, encoding='ascii') + b'\0'
        return NoteSection.build({  # type: ignore
            'namesz': len(b_name),
            'descsz': len(desc),
            'type': sec_type,
            'name': b_name,
            'desc': desc,
        })

    def _extract_bin_corefile(self, e_machine=ESPCoreDumpElfFile.EM_XTENSA):  # type: (Optional[int]) -> None
        """
        Creates core dump ELF file
        """
        coredump_data_struct = Struct(
            'tasks' / GreedyRange(
                AlignedStruct(
                    4,
                    'task_header' / TaskHeader,
                    'tcb' / Bytes(self.header.tcbsz),  # type: ignore
                    'stack' / Bytes(abs_(this.task_header.stack_top - this.task_header.stack_end)),  # type: ignore
                )
            ),
            'mem_seg_headers' / MemSegmentHeader[self.core_src.header.segs_num]  # type: ignore
        )
        core_elf = ESPCoreDumpElfFile(e_machine=e_machine)
        notes = b''
        core_dump_info_notes = b''
        task_info_notes = b''

        coredump_data = coredump_data_struct.parse(self.core_src.data)  # type: ignore
        for i, task in enumerate(coredump_data.tasks):
            stack_len_aligned = self._get_aligned_size(abs(task.task_header.stack_top - task.task_header.stack_end))
            task_status_kwargs = {
                'task_index': i,
                'task_flags': TASK_STATUS_CORRECT,
                'task_tcb_addr': task.task_header.tcb_addr,
                'task_stack_start': min(task.task_header.stack_top, task.task_header.stack_end),
                'task_stack_end': max(task.task_header.stack_top, task.task_header.stack_end),
                'task_stack_len': stack_len_aligned,
                'task_name': Padding(16).build({})  # currently we don't have task_name, keep it as padding
            }

            # Write TCB
            try:
                if self.target_methods.tcb_is_sane(task.task_header.tcb_addr, self.header.tcbsz):  # type: ignore
                    core_elf.add_segment(task.task_header.tcb_addr,
                                         task.tcb,
                                         ElfFile.PT_LOAD,
                                         ElfSegment.PF_R | ElfSegment.PF_W)
                elif task.task_header.tcb_addr and self.target_methods.addr_is_fake(task.task_header.tcb_addr):
                    task_status_kwargs['task_flags'] |= TASK_STATUS_TCB_CORRUPTED
            except ESPCoreDumpLoaderError as e:
                logging.warning('Skip TCB {} bytes @ 0x{:x}. (Reason: {})'
                                .format(self.header.tcbsz, task.task_header.tcb_addr, e))  # type: ignore

            # Write stack
            try:
                if self.target_methods.stack_is_sane(task_status_kwargs['task_stack_start'],
                                                     task_status_kwargs['task_stack_end']):
                    core_elf.add_segment(task_status_kwargs['task_stack_start'],
                                         task.stack,
                                         ElfFile.PT_LOAD,
                                         ElfSegment.PF_R | ElfSegment.PF_W)
                elif (task_status_kwargs['task_stack_start']
                      and self.target_methods.addr_is_fake(task_status_kwargs['task_stack_start'])):
                    task_status_kwargs['task_flags'] |= TASK_STATUS_TCB_CORRUPTED
                    core_elf.add_segment(task_status_kwargs['task_stack_start'],
                                         task.stack,
                                         ElfFile.PT_LOAD,
                                         ElfSegment.PF_R | ElfSegment.PF_W)
            except ESPCoreDumpLoaderError as e:
                logging.warning('Skip task\'s ({:x}) stack {} bytes @ 0x{:x}. (Reason: {})'
                                .format(task_status_kwargs['tcb_addr'],
                                        task_status_kwargs['stack_len_aligned'],
                                        task_status_kwargs['stack_base'],
                                        e))

            try:
                logging.debug('Stack start_end: 0x{:x} @ 0x{:x}'
                              .format(task.task_header.stack_top, task.task_header.stack_end))
                task_regs, extra_regs = self.target_methods.get_registers_from_stack(
                    task.stack,
                    task.task_header.stack_end > task.task_header.stack_top
                )
            except Exception as e:
                raise ESPCoreDumpLoaderError(str(e))

            task_info_notes += self._build_note_section('TASK_INFO',
                                                        ESPCoreDumpElfFile.PT_ESP_TASK_INFO,
                                                        EspTaskStatus.build(task_status_kwargs))
            notes += self._build_note_section('CORE',
                                              ElfFile.PT_LOAD,
                                              self.target_methods.build_prstatus_data(task.task_header.tcb_addr,
                                                                                      task_regs))

            if len(core_dump_info_notes) == 0:  # the first task is the crashed task
                core_dump_info_notes += self._build_note_section('ESP_CORE_DUMP_INFO',
                                                                 ESPCoreDumpElfFile.PT_ESP_INFO,
                                                                 Int32ul.build(self.header.ver))  # type: ignore
                _regs = [task.task_header.tcb_addr]

                # For xtensa, we need to put the exception registers into the extra info as well
                if e_machine == ESPCoreDumpElfFile.EM_XTENSA and extra_regs:
                    for reg_id in extra_regs:
                        _regs.extend([reg_id, extra_regs[reg_id]])

                core_dump_info_notes += self._build_note_section(
                    'EXTRA_INFO',
                    ESPCoreDumpElfFile.PT_ESP_EXTRA_INFO,
                    Int32ul[len(_regs)].build(_regs)
                )

        if self.dump_ver == self.BIN_V2:
            for header in coredump_data.mem_seg_headers:
                logging.debug('Read memory segment {} bytes @ 0x{:x}'.format(header.mem_sz, header.mem_start))
                core_elf.add_segment(header.mem_start, header.data, ElfFile.PT_LOAD, ElfSegment.PF_R | ElfSegment.PF_W)

        # add notes
        try:
            core_elf.add_segment(0, notes, ElfFile.PT_NOTE, 0)
        except ESPCoreDumpLoaderError as e:
            logging.warning('Skip NOTES segment {:d} bytes @ 0x{:x}. (Reason: {})'.format(len(notes), 0, e))
        # add core dump info notes
        try:
            core_elf.add_segment(0, core_dump_info_notes, ElfFile.PT_NOTE, 0)
        except ESPCoreDumpLoaderError as e:
            logging.warning('Skip core dump info NOTES segment {:d} bytes @ 0x{:x}. (Reason: {})'
                            .format(len(core_dump_info_notes), 0, e))
        try:
            core_elf.add_segment(0, task_info_notes, ElfFile.PT_NOTE, 0)
        except ESPCoreDumpLoaderError as e:
            logging.warning('Skip failed tasks info NOTES segment {:d} bytes @ 0x{:x}. (Reason: {})'
                            .format(len(task_info_notes), 0, e))
        # dump core ELF
        core_elf.e_type = ElfFile.ET_CORE
        core_elf.dump(self.core_elf_file)  # type: ignore


class ESPCoreDumpFlashLoader(EspCoreDumpLoader):
    ESP_COREDUMP_PART_TABLE_OFF = 0x8000

    def __init__(self, offset, target=None, port=None, baud=None, part_table_offset=0x8000):
        # type: (Optional[int], Optional[str], Optional[str], Optional[int], Optional[int]) -> None
        # TODO in next major release drop offset argument and use just parttool to find offset of coredump partition
        super(ESPCoreDumpFlashLoader, self).__init__()
        self.port = port
        self.baud = baud
        self.part_table_offset = part_table_offset

        self._get_core_src(offset, target)
        self.target = self._load_core_src()

    def _get_core_src(self, off, target=None):  # type: (Optional[int], Optional[str]) -> None
        """
        Loads core dump from flash using esptool
        """
        logging.info('Invoke esptool to read image.')
        self._invoke_esptool(off=off, target=target)

    def _invoke_esptool(self, off=None, target=None):  # type: (Optional[int], Optional[str]) -> None
        """
        Loads core dump from flash using elftool
        """
        if target is None:
            target = 'auto'
        tool_args = [sys.executable, '-m', 'esptool', '-c', target]
        if self.port:
            tool_args.extend(['-p', self.port])
        if self.baud:
            tool_args.extend(['-b', str(self.baud)])

        self.core_src_file = self._create_temp_file()
        try:
            (part_offset, part_size) = self._get_core_dump_partition_info()
            if not off:
                off = part_offset  # set default offset if not specified
                logging.warning('The core dump image offset is not specified. Use partition offset: 0x%x.', part_offset)
            if part_offset != off:
                logging.warning('Predefined image offset: %d does not match core dump partition offset: 0x%x', off,
                                part_offset)

            # Here we use V1 format to locate the size
            tool_args.extend(['read_flash', str(off), str(EspCoreDumpV1Header.sizeof())])
            tool_args.append(self.core_src_file)  # type: ignore

            # read core dump length
            et_out = subprocess.check_output(tool_args, stderr=subprocess.STDOUT)
            if et_out:
                logging.info(et_out.decode('utf-8'))

            header = EspCoreDumpV1Header.parse(open(self.core_src_file, 'rb').read())  # type: ignore
            if not header or not 0 < header.tot_len <= part_size:
                logging.error('Incorrect size of core dump image: {}, use partition size instead: {}'
                              .format(header.tot_len, part_size))
                coredump_len = part_size
            else:
                coredump_len = header.tot_len
            # set actual size of core dump image and read it from flash
            tool_args[-2] = str(coredump_len)
            et_out = subprocess.check_output(tool_args, stderr=subprocess.STDOUT)
            if et_out:
                logging.info(et_out.decode('utf-8'))
        except subprocess.CalledProcessError as e:
            raise ESPCoreDumpLoaderError(f'esptool script execution failed with error {e.returncode}, '
                                         f"failed command was: '{e.cmd}'",
                                         extra_output=e.output.decode('utf-8', 'ignore'))

    def _get_core_dump_partition_info(self):  # type: () -> Tuple[int, int]
        """
        Get core dump partition info using parttool
        """
        logging.info('Retrieving core dump partition offset and size...')
        part_off = self.part_table_offset or self.ESP_COREDUMP_PART_TABLE_OFF
        try:
            tool_args = [sys.executable, PARTTOOL_PY, '-q', '--partition-table-offset', str(part_off)]
            if self.port:
                tool_args.extend(['--port', self.port])
            invoke_args = tool_args + ['get_partition_info', '--partition-type', 'data',
                                       '--partition-subtype', 'coredump',
                                       '--info', 'offset', 'size']
            res = subprocess.check_output(invoke_args).strip()
            (offset_str, size_str) = res.rsplit(b'\n')[-1].split(b' ')
            size = int(size_str, 16)
            offset = int(offset_str, 16)
            logging.info('Core dump partition offset=%d, size=%d', offset, size)
        except subprocess.CalledProcessError as e:
            raise ESPCoreDumpLoaderError(f'parttool script execution failed with error {e.returncode}, '
                                         f"failed command was: '{' '.join(e.cmd)}'",
                                         extra_output=e.output.decode('utf-8', 'ignore'))
        return offset, size


class ESPCoreDumpFileLoader(EspCoreDumpLoader):
    def __init__(self, path, is_b64=False):  # type: (str, bool) -> None
        super(ESPCoreDumpFileLoader, self).__init__()
        self.is_b64 = is_b64

        self._get_core_src(path)
        self.target = self._load_core_src()

    def _get_core_src(self, path):  # type: (str) -> None
        """
        Loads core dump from (raw binary or base64-encoded) file
        """
        logging.debug('Load core dump from "%s", %s format', path, 'b64' if self.is_b64 else 'raw')
        if not self.is_b64:
            self.core_src_file = path
        else:
            self.core_src_file = self._create_temp_file()
            with open(self.core_src_file, 'wb') as fw:
                with open(path, 'rb') as fb64:
                    while True:
                        line = fb64.readline()
                        if len(line) == 0:
                            break
                        data = base64.standard_b64decode(line.rstrip(b'\r\n'))
                        fw.write(data)  # type: ignore



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#


from abc import ABC, abstractmethod
from importlib import import_module
from typing import Optional, Tuple

from .elf import ElfFile

__all__ = [
    'ElfFile', 'ESPCoreDumpLoaderError'
]

XTENSA_TARGETS = ['esp32', 'esp32s2', 'esp32s3']
RISCV_TARGETS = ['esp32c3', 'esp32c2', 'esp32c6', 'esp32h2', 'esp32p4', 'esp32c5', 'esp32c61']
SUPPORTED_TARGETS = XTENSA_TARGETS + RISCV_TARGETS


class ESPCoreDumpError(RuntimeError):
    pass


class ESPCoreDumpLoaderError(ESPCoreDumpError):
    def __init__(self, message, extra_output=None) -> None:
        super().__init__(message)
        self.extra_output = extra_output


class BaseArchMethodsMixin(ABC):  # type: ignore
    @staticmethod
    @abstractmethod
    def get_registers_from_stack(data, grows_down):
        # type: (bytes, bool) -> Tuple[list[int], Optional[dict[int, int]]]
        """
        Parse stack data, growing up stacks are not supported for now.
        :param data: stack data
        :param grows_down: stack grow direction
        :return: return tuple (regs, exception_regs)
        """
        pass

    @staticmethod
    @abstractmethod
    def build_prstatus_data(tcb_addr, task_regs):  # type: (int, list[int]) -> str
        """
        Build PrStatus note section
        :param tcb_addr: tcb addr
        :param task_regs: registers
        :return: str
        """
        pass


class BaseTargetMethods(BaseArchMethodsMixin, ABC):  # type: ignore
    UNKNOWN = 'unknown'
    TARGET = UNKNOWN

    COREDUMP_FAKE_STACK_START = 0x20000000
    COREDUMP_FAKE_STACK_LIMIT = 0x30000000
    COREDUMP_MAX_TASK_STACK_SIZE = 64 * 1024

    def __init__(self):  # type: () -> None
        if self.TARGET == self.UNKNOWN:
            raise ValueError('Please use the derived child-class with valid TARGET')

        self._set_attr_from_soc_header()

    def _set_attr_from_soc_header(self):  # type: () -> None
        module = import_module('esp_coredump.corefile.soc_headers.{}'.format(self.TARGET))
        for k, v in module.__dict__.items():
            if k.startswith('SOC_'):
                setattr(self, k, v)

    def _esp_ptr_in_dram(self, addr):  # type: (int) -> bool
        return self.SOC_DRAM_LOW <= addr < self.SOC_DRAM_HIGH  # type: ignore

    def _esp_ptr_in_iram(self, addr):  # type: (int) -> bool
        return self.SOC_IRAM_LOW <= addr < self.SOC_IRAM_HIGH  # type: ignore

    def _esp_ptr_in_rtc_slow(self, addr):  # type: (int) -> bool
        return self.SOC_RTC_DATA_LOW <= addr < self.SOC_RTC_DATA_HIGH  # type: ignore

    def _esp_ptr_in_rtc_dram_fast(self, addr):  # type: (int) -> bool
        return self.SOC_RTC_DRAM_LOW <= addr < self.SOC_RTC_DRAM_HIGH  # type: ignore

    def tcb_is_sane(self, tcb_addr, tcb_size):  # type: (int, int) -> bool
        for func in [self._esp_ptr_in_dram,
                     self._esp_ptr_in_iram,
                     self._esp_ptr_in_rtc_slow,
                     self._esp_ptr_in_rtc_dram_fast]:
            res = func(tcb_addr) and func(tcb_addr + tcb_size - 1)
            if res:
                return True
        return False

    def _esp_stack_ptr_in_dram(self, addr):  # type: (int) -> bool
        return not (addr < self.SOC_DRAM_LOW + 0x10  # type: ignore
                    or addr > self.SOC_DRAM_HIGH - 0x10  # type: ignore
                    or (addr & 0xF) != 0)

    def stack_is_sane(self, stack_start, stack_end):  # type: (int, int) -> bool
        return (self._esp_stack_ptr_in_dram(stack_start)
                and self._esp_ptr_in_dram(stack_end)
                and stack_start < stack_end
                and (stack_end - stack_start) < self.COREDUMP_MAX_TASK_STACK_SIZE)

    def addr_is_fake(self, addr):  # type: (int) -> bool
        return (self.COREDUMP_FAKE_STACK_START <= addr < self.COREDUMP_FAKE_STACK_LIMIT
                or addr > 2 ** 31 - 1)



SOC_IRAM_LOW = 0x4037c000
SOC_IRAM_HIGH = 0x403c0000
SOC_DRAM_LOW = 0x3fca0000
SOC_DRAM_HIGH = 0x3fce0000
SOC_RTC_DATA_LOW = 0xffffffff
SOC_RTC_DATA_HIGH = 0xffffffff
SOC_RTC_DRAM_LOW = 0xffffffff
SOC_RTC_DRAM_HIGH = 0xffffffff



SOC_IRAM_LOW = 0x4ff00000
SOC_IRAM_HIGH = 0x4ffc0000
SOC_DRAM_LOW = 0x4ff00000
SOC_DRAM_HIGH = 0x4ffc0000
SOC_RTC_IRAM_LOW = 0x50108000
SOC_RTC_IRAM_HIGH = 0x50110000
SOC_RTC_DRAM_LOW = 0x50108000
SOC_RTC_DRAM_HIGH = 0x50110000
SOC_RTC_DATA_LOW = 0x50108000
SOC_RTC_DATA_HIGH = 0x50110000



#
# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import argparse
import os

from esp_coredump import __version__

from .corefile import SUPPORTED_TARGETS
from .corefile.gdb import DEFAULT_GDB_TIMEOUT_SEC

try:
    # esptool>=4.0
    from esptool.loader import ESPLoader
except (AttributeError, ModuleNotFoundError):
    # esptool<4.0
    from esptool import ESPLoader


def arg_auto_int(x):
    return int(x, 0)


parser = argparse.ArgumentParser(description=f'espcoredump.py v{__version__} - ESP32 Core Dump Utility')
parser.add_argument('--chip', default=os.environ.get('ESPTOOL_CHIP', 'auto'),
                    choices=['auto'] + SUPPORTED_TARGETS,
                    help='Target chip type')
parser.add_argument('--chip-rev', type=int,
                    help='Target chip revision')
parser.add_argument('--port', '-p', default=os.environ.get('ESPTOOL_PORT'),
                    help='Serial port device')
parser.add_argument('--baud', '-b', type=int,
                    default=os.environ.get('ESPTOOL_BAUD', ESPLoader.ESP_ROM_BAUD),
                    help='Serial port baud rate used when flashing/reading')
parser.add_argument('--gdb-timeout-sec', type=int, default=DEFAULT_GDB_TIMEOUT_SEC,
                    help='Overwrite the default internal delay for gdb responses')
parser.add_argument('--version', action='version',
                    version=f'espcoredump.py v{__version__}')

common_args = argparse.ArgumentParser(add_help=False)
common_args.add_argument('--debug', '-d', type=int, default=3,
                         help='Log level (0..3)')  # TODO: move this option to global args in next major release
common_args.add_argument('--gdb', '-g',
                         help='Path to gdb')
common_args.add_argument('--extra-gdbinit-file', '-ex',
                         help='Path to additional gdbinit file')
common_args.add_argument('--core', '-c',
                         help='Path to core dump file (if skipped core dump will be read from flash)')
common_args.add_argument('--core-format', '-t', choices=['auto', 'b64', 'elf', 'raw'], default='auto',
                         help='File specified with "-c" is an ELF (elf), raw (raw) or base64-encoded (b64) binary. '
                         'For autodetection based on file header use "auto".')
common_args.add_argument('--off', '-o', type=arg_auto_int,
                         help='Offset of coredump partition in flash (type "idf.py partition-table" to see).')
common_args.add_argument('--parttable-off', type=arg_auto_int,
                         help='Offset of the partition table in flash.')
common_args.add_argument('--save-core', '-s',
                         help='Save core to file. Otherwise temporary core file will be deleted. '
                              'Does not work with "-c"', )
common_args.add_argument('--rom-elf', '-r',
                         help='Path to ROM ELF file. Will use "<target>_rom.elf" if not specified')
common_args.add_argument('prog', help='Path to program\'s ELF binary')

operations = parser.add_subparsers(dest='operation', required=True, description='Operation to perform')

operations.add_parser('dbg_corefile', parents=[common_args],
                      help='Starts GDB debugging session with specified corefile')

info_coredump = operations.add_parser('info_corefile', parents=[common_args],
                                      help='Print core dump info from file')
info_coredump.add_argument('--print-mem', '-m', action='store_true',
                           help='Print memory dump')



# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import base64
import contextlib
import io
import os
import subprocess
import sys
import time

import pytest

try:
    from esp_coredump import CoreDump
    from esp_coredump.corefile import ESPCoreDumpLoaderError
    from esp_coredump.corefile.elf import ESPCoreDumpElfFile
    from esp_coredump.corefile.loader import ESPCoreDumpFileLoader
except ImportError:
    raise ModuleNotFoundError('No module named "esp_coredump" please install esp_coredump by running '
                              '"python -m pip install esp-coredump"')

SUPPORTED_TARGET = ['esp32', 'esp32c3', 'esp32p4']
COREDUMP_FILE_NAME = 'coredump'
COREDUMP_BIN_FILE_NAME = 'coredump_bin'

TEST_DIR_ABS_PATH = os.path.dirname(__file__)
ESP_PROG_DIR = os.path.join(TEST_DIR_ABS_PATH, 'test_apps', 'built_apps')


@pytest.fixture(scope='session')
def coverage_run():
    """Run with coverage reporting if available"""
    try:
        import coverage  # noqa: F401
        yield ['coverage', 'run', '--parallel-mode']
    except ImportError:
        yield [sys.executable]


def get_coredump_kwargs(core_ext: str, target: str, save_core: bool = False, auto_format: bool = False, bin_fmt: bool = False):
    core_format = 'auto' if auto_format else 'raw' if core_ext == 'bin' else core_ext
    coredump_filename = COREDUMP_BIN_FILE_NAME if bin_fmt else COREDUMP_FILE_NAME
    kwargs = {
        'gdb_timeout_sec': 5,
        'chip': 'auto',
        'print_mem': True,
        'core_format': core_format,
        'core': os.path.join(TEST_DIR_ABS_PATH, target, f'{coredump_filename}.{core_ext}'),
        'save_core': os.path.join(TEST_DIR_ABS_PATH, target, f'{coredump_filename}.elf') if save_core else None,
        'prog': os.path.join(ESP_PROG_DIR, f'{target + ("_bin" if bin_fmt else "")}.elf'),
    }
    return kwargs


def get_expected_output(target: str, bin_fmt: bool = False):
    expected_output_file = os.path.join(TEST_DIR_ABS_PATH, target, 'expected_output' + ('_bin' if bin_fmt else ''))
    with open(expected_output_file) as file:
        output = file.read()
    return output


def get_output(core_ext: str, target: str, save_core: bool = False, auto_format: bool = False, bin_fmt: bool = False):
    kwargs = get_coredump_kwargs(core_ext=core_ext, save_core=save_core, target=target, auto_format=auto_format, bin_fmt=bin_fmt)
    coredump = CoreDump(**kwargs)
    output_file = os.path.join(TEST_DIR_ABS_PATH, target, f'output_from_{("bin_" if bin_fmt else "") + core_ext}')
    with io.StringIO() as buffer, contextlib.redirect_stdout(buffer):
        coredump.info_corefile()
        output = buffer.getvalue()

    with open(output_file, 'w') as f:
        f.write(output)
    return output


def decode_from_b64_to_bin(target):
    b64_file_path = os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.b64')
    bin_file_path = os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.bin')
    with open(b64_file_path, 'rb') as fb64, open(bin_file_path, 'wb') as fw:
        for line in fb64:
            data = base64.standard_b64decode(line.rstrip(b'\r\n'))
            fw.write(data)


class TestESPCoreDumpDecode:

    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_coredump_decode_from_cli(self, coverage_run, target):
        proc = subprocess.run(
            coverage_run + [
                '-m', 'esp_coredump', 'info_corefile',
                '--core', os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.b64'),
                '--save-core', os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.elf'),
                '--print-mem',
                os.path.join(ESP_PROG_DIR, f'{target}.elf'),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        expected_output = get_expected_output(target)
        out = proc.stdout.decode().split('\n', 1)[1]  # remove first line with version
        assert expected_output == out

    @pytest.mark.parametrize('target', ['esp32', 'esp32c3'])
    def test_bin_coredump_decode(self, target):
        """Built with CONFIG_ESP_COREDUMP_DATA_FORMAT_BIN=y (instead of default ELF)"""
        output = get_output(core_ext='b64', target=target, bin_fmt=True)
        expected_output = get_expected_output(target, bin_fmt=True)
        assert expected_output == output

    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_coredump_decode_from_b64(self, target):
        output = get_output(core_ext='b64', save_core=True, target=target)
        expected_output = get_expected_output(target)
        assert expected_output == output

    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_coredump_decode_from_elf(self, target):
        output = get_output(core_ext='elf', target=target)
        expected_output = get_expected_output(target)
        assert expected_output == output

    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_coredump_decode_from_bin(self, target):
        decode_from_b64_to_bin(target)
        output = get_output(core_ext='bin', target=target)
        expected_output = get_expected_output(target)
        assert expected_output == output

    @pytest.mark.parametrize('format', ['bin', 'elf', 'b64'])
    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_coredump_decode_auto_format(self, target, format):
        # make sure that .elf and .bin inputs are created
        get_output(core_ext='b64', save_core=True, target=target)
        decode_from_b64_to_bin(target)
        output = get_output(core_ext=format, target=target, auto_format=True)
        expected_output = get_expected_output(target)
        assert expected_output == output


class TestESPCoreDumpElfFile:
    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_read_elf(self, target):
        elf = ESPCoreDumpElfFile(os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.elf'))
        assert elf.load_segments is not None
        assert elf.note_segments is not None


class TestESPCoreDumpFileLoader:
    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_load_wrong_encode_core_bin(self, target):
        with pytest.raises(ESPCoreDumpLoaderError):
            ESPCoreDumpFileLoader(path=os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.b64'), is_b64=False)

    @pytest.mark.parametrize('target', SUPPORTED_TARGET)
    def test_create_corefile(self, target):
        loader = ESPCoreDumpFileLoader(path=os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.b64'), is_b64=True)
        loader.create_corefile()
        assert os.path.exists(loader.core_elf_file)


class TestDebugCoredump:
    def test_dbg_corefile(self, coverage_run):
        target = 'esp32'
        proc = subprocess.Popen(
            coverage_run + [
                '-m', 'esp_coredump', 'dbg_corefile',
                '--core', os.path.join(TEST_DIR_ABS_PATH, target, f'{COREDUMP_FILE_NAME}.b64'),
                os.path.join(ESP_PROG_DIR, f'{target}.elf'),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            bufsize=0,
        )
        time.sleep(2)  # wait for gdb to start
        proc.stdin.write(b'\n')
        proc.stdin.write(b'y\n')
        stdout, _ = proc.communicate(b'q\n')
        proc.wait(timeout=3)
        stdout = stdout.decode()
        assert stdout is not None
        assert 'GNU gdb (esp-gdb)' in stdout
        assert f'Reading symbols from {os.path.join(ESP_PROG_DIR, f"{target}.elf")}' in stdout
        assert 'Done!' in stdout



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Optional, Tuple

from construct import Int16ul, Int32ul, Padding, Struct

from . import BaseArchMethodsMixin, BaseTargetMethods, ESPCoreDumpLoaderError

RISCV_GP_REGS_COUNT = 32
PRSTATUS_SIZE = 204
PRSTATUS_OFFSET_PR_CURSIG = 12
PRSTATUS_OFFSET_PR_PID = 24
PRSTATUS_OFFSET_PR_REG = 72
ELF_GREGSET_T_SIZE = 128

PrStruct = Struct(
    Padding(PRSTATUS_OFFSET_PR_CURSIG),
    'pr_cursig' / Int16ul,
    Padding(PRSTATUS_OFFSET_PR_PID - PRSTATUS_OFFSET_PR_CURSIG - Int16ul.sizeof()),
    'pr_pid' / Int32ul,
    Padding(PRSTATUS_OFFSET_PR_REG - PRSTATUS_OFFSET_PR_PID - Int32ul.sizeof()),
    'regs' / Int32ul[RISCV_GP_REGS_COUNT],
    Padding(PRSTATUS_SIZE - PRSTATUS_OFFSET_PR_REG - ELF_GREGSET_T_SIZE)
)


class RiscvMethodsMixin(BaseArchMethodsMixin):
    @staticmethod
    def get_registers_from_stack(data, grows_down):
        # type: (bytes, bool) -> Tuple[list[int], Optional[dict[int, int]]]
        regs = Int32ul[RISCV_GP_REGS_COUNT].parse(data)
        if not grows_down:
            raise ESPCoreDumpLoaderError('Growing up stacks are not supported for now!')
        return regs, None

    @staticmethod
    def build_prstatus_data(tcb_addr, task_regs):  # type: (int, list[int]) -> Any
        return PrStruct.build({
            'pr_cursig': 0,
            'pr_pid': tcb_addr,
            'regs': task_regs,
        })


class Esp32C3Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32c3'


class Esp32C2Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32c2'


class Esp32H2Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32h2'


class Esp32C6Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32c6'


class Esp32P4Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32p4'


class Esp32C5Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32c5'


class Esp32C61Methods(BaseTargetMethods, RiscvMethodsMixin):
    TARGET = 'esp32c61'



#!/usr/bin/env python
#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#
# ESP-IDF Core Dump Utility
#
import os
import subprocess
import sys
import textwrap
from contextlib import contextmanager
from shutil import copyfile, which
from typing import Optional, Tuple, Union

import serial

from .tools import load_json_from_file

try:
    # esptool>=4.0
    from esptool.cmds import detect_chip
    from esptool.loader import ESPLoader
except (AttributeError, ModuleNotFoundError):
    # esptool<4.0
    from esptool import ESPLoader
    detect_chip = ESPLoader.detect_chip

from construct import Container, GreedyRange, Int32ul, ListContainer, Struct

from .corefile import RISCV_TARGETS, SUPPORTED_TARGETS, XTENSA_TARGETS, xtensa
from .corefile.elf import (TASK_STATUS_CORRECT, ElfFile, ElfSegment,
                           ESPCoreDumpElfFile, EspTaskStatus)
from .corefile.gdb import DEFAULT_GDB_TIMEOUT_SEC, EspGDB
from .corefile.loader import (ESPCoreDumpFileLoader, ESPCoreDumpFlashLoader,
                              ESPCoreDumpLoaderError, EspCoreDumpVersion,
                              get_core_file_format)

IDF_PATH = os.getenv('IDF_PATH', '')
ESP_ROM_ELF_DIR = os.getenv('ESP_ROM_ELF_DIR')
# 'tools/idf_py_actions/roms.json' is used for compatibility with ESP-IDF before v5.5, when the file was moved
ROMS_JSON = [os.path.join(IDF_PATH, 'components', 'esp_rom', 'roms.json'), os.path.join(IDF_PATH, 'tools', 'idf_py_actions', 'roms.json')]

MORE_INFO_MSG = 'Read more: https://github.com/espressif/esp-coredump/blob/master/README.md#installation'
GDB_NOT_FOUND_ERROR = (
    f'GDB executable not found. Please install GDB or set up ESP-IDF to complete the action. '
    f'{MORE_INFO_MSG}'
)
IDF_SETUP_ERROR = f'Please set up ESP-IDF to complete the action. {MORE_INFO_MSG}'

RETRY_ATTEMPTS = 3

XTENSA_ISR_CTX_IDX = 37
RISCV_ISR_CTX_IDX = 1

if os.name == 'nt':
    CLOSE_FDS = False
else:
    CLOSE_FDS = True


class CoreDump:
    def __init__(self,
                 baud: Optional[int] = int(os.environ.get('ESPTOOL_BAUD', ESPLoader.ESP_ROM_BAUD)),
                 chip: str = os.environ.get('ESPTOOL_CHIP', 'auto'),
                 core_format: str = 'auto',
                 port: Optional[str] = os.environ.get('ESPTOOL_PORT'),
                 gdb_timeout_sec: int = DEFAULT_GDB_TIMEOUT_SEC,
                 core: Optional[str] = None,
                 chip_rev: Optional[int] = None,
                 gdb: Optional[str] = None,
                 extra_gdbinit_file: Optional[str] = None,
                 off: Optional[int] = None,
                 parttable_off: Optional[int] = None,
                 prog: Optional[str] = None,
                 print_mem: Optional[str] = None,
                 rom_elf: Optional[str] = None,
                 save_core: Optional[str] = None,
                 ):
        if prog is None:
            raise ValueError("Path to program\'s ELF binary is not provided")

        self.baud = baud
        self.chip = chip
        self.core = core
        self.chip_rev = chip_rev
        self.core_format = get_core_file_format(core) if core and core_format == 'auto' else core_format
        self.gdb = gdb
        self.gdb_timeout_sec = gdb_timeout_sec
        self.extra_gdbinit_file = extra_gdbinit_file
        self.coredump_off = off
        self.parttable_off = parttable_off
        self.prog = prog
        self.port = port
        self.print_mem = print_mem
        self.rom_elf = rom_elf
        self.save_core = save_core

    @staticmethod
    def load_aux_elf(elf_path):  # type: (str) -> str
        """
        Loads auxiliary ELF file and composes GDB command to read its symbols.
        """
        sym_cmd = ''
        if os.path.exists(elf_path):
            elf = ElfFile(elf_path)
            if os.name == 'nt':
                elf_path = elf_path.replace('\\', '/')
            for s in elf.sections:
                if s.name == '.text':
                    sym_cmd = f'add-symbol-file {elf_path} {s.addr:#x}'
                    break
        return sym_cmd

    def extract_chip_rev_from_elf(self):
        if not os.path.exists(self.core):
            raise FileNotFoundError(f"Provided ELF file {self.core} is not found or doesn't exist")
        elf = ElfFile(elf_path=self.core)
        chip_rev = None
        for s in elf.note_segments:
            for n in s.note_secs:
                if b'ESP_CHIP_REV' in n.name:
                    chip_rev = int.from_bytes(n.desc, 'little')
                    break

        return chip_rev

    def get_core_header_info_dict(self, e_machine=ESPCoreDumpElfFile.EM_XTENSA):
        loader = None  # type: Union[ESPCoreDumpFlashLoader, ESPCoreDumpFileLoader, None]
        core_dump_info_map = {
            'core_elf_path': None,
            'target': None,
            'temp_files': None,
            'chip_rev': None,
        }

        if not self.core:
            # Core file not specified, try to read core dump from flash.
            if not IDF_PATH:
                print(IDF_SETUP_ERROR)
                sys.exit(1)
            loader = ESPCoreDumpFlashLoader(self.coredump_off, port=self.port, baud=self.baud, part_table_offset=self.parttable_off)
        elif self.core_format != 'elf':
            # Core file specified, but not yet in ELF format. Convert it from raw or base64 into ELF.
            loader = ESPCoreDumpFileLoader(self.core, self.core_format == 'b64')
        else:
            # Core file is already in the ELF format
            core_dump_info_map['core_elf_path'] = self.core
            chip_rev = self.extract_chip_rev_from_elf()

            if self.chip_rev is not None and chip_rev != self.chip_rev:
                print('Provided chip revision does not match the one extracted from the provided coredump elf file.',
                      file=sys.stderr)
                exit(1)

            core_dump_info_map['chip_rev'] = chip_rev

        # Load/convert the core file
        if loader:
            loader.create_corefile(exe_name=self.prog, e_machine=e_machine)
            core_dump_info_map['core_elf_path'] = loader.core_elf_file
            if self.save_core:
                # We got asked to save the core file, make a copy
                copyfile(loader.core_elf_file, self.save_core)
            core_dump_info_map['target'] = loader.target
            core_dump_info_map['chip_rev'] = loader.chip_rev
            core_dump_info_map['temp_files'] = loader.temp_files

        return core_dump_info_map

    def get_chip_version(self):  # type: () -> Optional[int]
        for segment in self.core_elf.note_segments:
            for sec in segment.note_secs:
                if sec.type == ESPCoreDumpElfFile.PT_ESP_INFO:
                    ver_bytes = sec.desc[:4]
                    return int((ver_bytes[3] << 8) | ver_bytes[2])
        return None

    def get_target(self):  # type: () -> str

        if self.chip != 'auto':
            return self.chip  # type: ignore

        chip_version = self.get_chip_version()
        if chip_version is not None:
            if chip_version == EspCoreDumpVersion.ESP32:
                return 'esp32'

            if chip_version == EspCoreDumpVersion.ESP32S2:
                return 'esp32s2'

            if chip_version == EspCoreDumpVersion.ESP32S3:
                return 'esp32s3'

            if chip_version == EspCoreDumpVersion.ESP32C3:
                return 'esp32c3'

            if chip_version == EspCoreDumpVersion.ESP32C2:
                return 'esp32c2'

            if chip_version == EspCoreDumpVersion.ESP32C6:
                return 'esp32c6'

            if chip_version == EspCoreDumpVersion.ESP32H2:
                return 'esp32h2'

            if chip_version == EspCoreDumpVersion.ESP32P4:
                return 'esp32p4'

            if chip_version == EspCoreDumpVersion.ESP32C5:
                return 'esp32c5'

            if chip_version == EspCoreDumpVersion.ESP32C61:
                return 'esp32c61'

        target = None
        try:
            inst = detect_chip(self.port, self.baud)
        except serial.serialutil.SerialException:
            print('Unable to identify the chip type. '
                  'Please use the --chip option to specify the chip type or '
                  'connect the board and provide the --port option to have the chip type determined automatically.',
                  file=sys.stderr)
            exit(1)
        else:
            target = inst.CHIP_NAME.lower().replace('-', '')

        return target  # type: ignore

    def get_gdb_path(self, target):  # type: (str) -> Optional[str]
        if self.gdb:
            return self.gdb  # type: ignore

        if target in XTENSA_TARGETS:
            # For some reason, xtensa-esp32s2-elf-gdb will report some issue.
            # Use xtensa-esp32-elf-gdb instead.
            gdb_path = 'xtensa-esp32-elf-gdb'
        elif target in RISCV_TARGETS:
            gdb_path = 'riscv32-esp-elf-gdb'
        else:
            raise ValueError(f'Invalid value: {target}. For now we only support {SUPPORTED_TARGETS}')
        if which(gdb_path) is None:
            return None

        return gdb_path

    def get_gdb_args(self, target, core_elf_path, chip_rev, is_dbg_mode=False):
        gdb_tool = self.get_gdb_path(target)
        if not gdb_tool:
            print(GDB_NOT_FOUND_ERROR)
            sys.exit(1)

        rom_elf_path = self.get_rom_elf_path(target=target, chip_rev=chip_rev)
        rom_sym_cmd = self.load_aux_elf(rom_elf_path)

        gdb_args = [gdb_tool]

        if self.extra_gdbinit_file:
            if not os.path.isfile(self.extra_gdbinit_file):
                raise ValueError(f'{self.extra_gdbinit_file} does not exist')
            gdb_args.append('-x={}'.format(self.extra_gdbinit_file))
        else:
            gdb_args.append('--nx')  # ignore .gdbinit

        if not is_dbg_mode:
            gdb_args.append('--nw')  # suppress the GUI interface
            gdb_args.append('--quiet')  # inhibit dumping info at start-up
            gdb_args.append('--interpreter=mi2')  # use GDB/MI v2

        gdb_args.append('--core={}'.format(core_elf_path))  # core file
        if rom_sym_cmd:
            gdb_args += ['-ex', rom_sym_cmd]
        gdb_args.append(self.prog)

        return gdb_args

    def get_rom_elf_path(self, chip_rev, target):  # type: (Optional[int], str) -> str
        if self.rom_elf:
            return self.rom_elf

        if chip_rev is None:
            return ''

        rom_file_path = ''

        if not IDF_PATH or not ESP_ROM_ELF_DIR:
            print("The ROM ELF file won't be loaded automatically since you are running the utility out of IDF.")
            return rom_file_path

        target_roms = None
        for rom_json_path in ROMS_JSON:
            try:
                roms_json = load_json_from_file(rom_json_path)
                target_roms = roms_json.get(target, [])
                break
            except FileNotFoundError:
                continue

        if not target_roms:
            print("The ROM ELF file won't load automatically since it was not found for the provided chip type.")
            return rom_file_path

        index = len(target_roms)
        for idx in range(len(target_roms)):
            if target_roms[idx].get('rev') == chip_rev:
                index = idx
                break
        if index < len(target_roms):
            chip_rev_from_json = target_roms[index]['rev']
            rom_elf_file_name = f'{target}_rev{chip_rev_from_json}_rom.elf'
            rom_file_path = os.path.join(ESP_ROM_ELF_DIR, rom_elf_file_name)
        else:
            print("The ROM ELF file won't load automatically since it was not found for the provided chip type.")

        return rom_file_path

    def get_task_info_extra_note_tuple(self):  # type: () -> Tuple[Optional[list[str]], Optional[Container]]
        extra_note = None
        task_info = []
        for note_seg in self.core_elf.note_segments:
            for note_sec in note_seg.note_secs:
                if note_sec.type == ESPCoreDumpElfFile.PT_ESP_EXTRA_INFO:
                    extra_note = note_sec
                if note_sec.type == ESPCoreDumpElfFile.PT_ESP_TASK_INFO:
                    task_info_struct = EspTaskStatus.parse(note_sec.desc)
                    task_info.append(task_info_struct)
        return task_info, extra_note

    def get_panic_details(self):
        for note_seg in self.core_elf.note_segments:
            for note_sec in note_seg.note_secs:
                if note_sec.type == ESPCoreDumpElfFile.PT_ESP_PANIC_DETAILS:
                    return note_sec
        return None

    def print_crashed_task_info(self, marker):  # type: (Optional[int]) -> None
        if marker == ESPCoreDumpElfFile.CURR_TASK_MARKER:
            print('\nCrashed task has been skipped.')
        else:
            task_name = self.gdb_esp.get_freertos_task_name(marker)
            print("\nCrashed task handle: 0x%x, name: '%s', GDB name: 'process %d'"
                  % (marker, task_name, marker))  # type: ignore

    def print_threads_info(self, task_info):  # type: (Optional[list[Container]]) -> None
        print(self.gdb_esp.run_cmd('info threads'))
        # THREADS STACKS
        for attempt in range(1, RETRY_ATTEMPTS + 1):
            threads, _ = self.gdb_esp.get_thread_info(attempt * DEFAULT_GDB_TIMEOUT_SEC)
            if threads:
                break

            print('Retrying reading threads information...')

        if not threads:
            print('\nThe threads information for the current task could not be retrieved. '
                  'Please try running this command again with --gdb-timeout-sec option to increase '
                  f'the default value of internal delay for gdb responses (Default: {DEFAULT_GDB_TIMEOUT_SEC})')
            return

        print('\n')
        print('       TCB             NAME PRIO C/B  STACK USED/FREE')
        print('---------- ---------------- -------- ----------------')

        thread_dict = {}
        for thr in threads:
            thr_id = int(thr['id'])
            tcb_addr = self.gdb_esp.gdb2freertos_thread_id(thr['target-id'])
            task_name = self.gdb_esp.get_freertos_task_name(tcb_addr)
            try:
                pxEndOfStack = int(self.gdb_esp.parse_tcb_variable(tcb_addr, 'pxEndOfStack'), 16)
                pxTopOfStack = int(self.gdb_esp.parse_tcb_variable(tcb_addr, 'pxTopOfStack'), 16)
                pxStack = int(self.gdb_esp.parse_tcb_variable(tcb_addr, 'pxStack'), 16)
                uxPriority = int(self.gdb_esp.parse_tcb_variable(tcb_addr, 'uxPriority'), 16)
                uxBasePriority = int(self.gdb_esp.parse_tcb_variable(tcb_addr, 'uxBasePriority'), 16)
            except ValueError:
                pxEndOfStack = pxTopOfStack = pxStack = uxPriority = uxBasePriority = 0

            thread_dict[thr_id] = {'tcb_addr': tcb_addr, 'task_name': task_name}
            ftcb_addr = '0x{:x}'.format(tcb_addr)
            if pxStack == 0:
                print(f'{ftcb_addr:>10} Corrupted TCB data')
            else:
                fpriority = '{}/{}'.format(uxPriority, uxBasePriority)
                fstack_usage = '{}/{}'.format(abs(pxEndOfStack - pxTopOfStack), abs(pxStack - pxTopOfStack))
                print(f'{ftcb_addr:>10}{task_name:>17}{fpriority:>9}{fstack_usage:>17}')

        for thr_id, value in thread_dict.items():
            tcb_addr = value['tcb_addr']
            task_index = thr_id - 1
            task_name = value['task_name']
            self.gdb_esp.switch_thread(thr_id)
            print('\n==================== THREAD {} (TCB: 0x{:x}, name: \'{}\') ====================='
                  .format(thr_id, tcb_addr, task_name))

            print(self.gdb_esp.run_cmd('bt'))
            if task_info and task_info[task_index].task_flags != TASK_STATUS_CORRECT:
                print("The task '%s' is corrupted." % thr_id)
                print('Task #%d info: flags, tcb, stack (%x, %x, %x).' % (task_info[task_index].task_index,
                                                                          task_info[task_index].task_flags,
                                                                          task_info[task_index].task_tcb_addr,
                                                                          task_info[task_index].task_stack_start))

    def print_current_thread_registers(self, extra_note, extra_info):
        # type: (Optional[Container], Optional[ListContainer]) -> None
        if self.exe_elf.e_machine == ESPCoreDumpElfFile.EM_XTENSA:
            if extra_note and extra_info:
                xtensa.print_exc_regs_info(extra_info)
            else:
                print('Exception registers have not been found!')
        print(self.gdb_esp.run_cmd('info registers'))

    def print_isr_context(self, extra_info):
        if self.exe_elf.e_machine == ESPCoreDumpElfFile.EM_XTENSA:
            isr_ctx_idx = XTENSA_ISR_CTX_IDX
        else:
            isr_ctx_idx = RISCV_ISR_CTX_IDX
        if (len(extra_info) < isr_ctx_idx + 1):
            # no information
            return
        if extra_info[isr_ctx_idx]:
            print('Crash happened in the interrupt context')
        else:
            print('Crashed task is not in the interrupt context')

    def print_current_thread_stack(self, task_info):  # type: (Optional[list[Container]]) -> None
        print(self.gdb_esp.run_cmd('bt'))
        if task_info and task_info[0].task_flags != TASK_STATUS_CORRECT:
            print('The current crashed task is corrupted.')
            print('Task #%d info: flags, tcb, stack (%x, %x, %x).' % (task_info[0].task_index,
                                                                      task_info[0].task_flags,
                                                                      task_info[0].task_tcb_addr,
                                                                      task_info[0].task_stack_start))

    def print_all_memory_regions(self):  # type: () -> None
        print('Name   Address   Size   Attrs')
        core_segs = self.core_elf.load_segments
        merged_segs = []
        for sec in self.exe_elf.sections:
            merged = False
            for seg in core_segs:
                if seg.addr <= sec.addr <= seg.addr + len(seg.data):
                    # sec:    |XXXXXXXXXX|
                    # seg: |...XXX.............|
                    seg_addr = seg.addr
                    if seg.addr + len(seg.data) <= sec.addr + len(sec.data):
                        # sec:        |XXXXXXXXXX|
                        # seg:    |XXXXXXXXXXX...|
                        # merged: |XXXXXXXXXXXXXX|
                        seg_len = len(sec.data) + (sec.addr - seg.addr)
                    else:
                        # sec:        |XXXXXXXXXX|
                        # seg:    |XXXXXXXXXXXXXXXXX|
                        # merged: |XXXXXXXXXXXXXXXXX|
                        seg_len = len(seg.data)
                    merged_segs.append((sec.name, seg_addr, seg_len, sec.attr_str(), True))
                    core_segs.remove(seg)
                    merged = True
                elif sec.addr <= seg.addr <= sec.addr + len(sec.data):
                    # sec:  |XXXXXXXXXX|
                    # seg:  |...XXX.............|
                    seg_addr = sec.addr
                    if (seg.addr + len(seg.data)) >= (sec.addr + len(sec.data)):
                        # sec:    |XXXXXXXXXX|
                        # seg:    |..XXXXXXXXXXX|
                        # merged: |XXXXXXXXXXXXX|
                        seg_len = len(sec.data) + (seg.addr + len(seg.data)) - (sec.addr + len(sec.data))
                    else:
                        # sec:    |XXXXXXXXXX|
                        # seg:      |XXXXXX|
                        # merged: |XXXXXXXXXX|
                        seg_len = len(sec.data)
                    merged_segs.append((sec.name, seg_addr, seg_len, sec.attr_str(), True))
                    core_segs.remove(seg)
                    merged = True

            if not merged:
                merged_segs.append((sec.name, sec.addr, len(sec.data), sec.attr_str(), False))

        for ms in merged_segs:
            print('%s 0x%x 0x%x %s' % (ms[0], ms[1], ms[2], ms[3]))

        for cs in core_segs:
            # core dump exec segments are from ROM, other are belong to tasks (TCB or stack)
            if cs.flags & ElfSegment.PF_X:
                seg_name = 'rom.text'
            else:
                seg_name = 'tasks.data'
            print('.coredump.%s 0x%x 0x%x %s' % (seg_name, cs.addr, len(cs.data), cs.attr_str()))

    def print_core_dump_memory_contents(self):  # type: () -> None
        for cs in self.core_elf.load_segments:
            # core dump exec segments are from ROM, other are belong to tasks (TCB or stack)
            if cs.flags & ElfSegment.PF_X:
                seg_name = 'rom.text'
            else:
                seg_name = 'tasks.data'
            print('.coredump.%s 0x%x 0x%x %s' % (seg_name, cs.addr, len(cs.data), cs.attr_str()))
            print(self.gdb_esp.run_cmd('x/%dx 0x%x' % (len(cs.data) // 4, cs.addr)))

    def verify_target(self, core_header_info_dict):
        target = core_header_info_dict.get('target')
        if target is None:
            target = self.get_target()
            core_header_info_dict['target'] = target
        return target

    @contextmanager
    def _handle_coredump_loader_error(self):
        try:
            yield
        except ESPCoreDumpLoaderError as e:
            print(f'Failed to load core dump: {e}', file=sys.stderr)
            if e.extra_output:
                print('', file=sys.stderr)
                print('┌────── Additional information about the error: ', file=sys.stderr)
                print('│   ', file=sys.stderr)
                print(textwrap.indent(e.extra_output, '│   '), file=sys.stderr)
                print('│   ', file=sys.stderr)
                print('└────── end of additional information about the error.', file=sys.stderr)
            raise SystemExit(1)

    def dbg_corefile(self):  # type: () -> Optional[list[str]]
        """
        Command to load core dump from file or flash and run GDB debug session with it
        """
        exe_elf = ESPCoreDumpElfFile(self.prog)
        with self._handle_coredump_loader_error():
            core_header_info_dict = self.get_core_header_info_dict(e_machine=exe_elf.e_machine)
            self.core_elf = ESPCoreDumpElfFile(core_header_info_dict['core_elf_path'])

        temp_files = core_header_info_dict.pop('temp_files')
        self.chip = self.verify_target(core_header_info_dict)

        gdb_args = self.get_gdb_args(is_dbg_mode=True, **core_header_info_dict)

        p = subprocess.Popen(bufsize=0,
                             args=gdb_args,
                             stdin=None, stdout=None, stderr=None,
                             close_fds=CLOSE_FDS)
        p.wait()
        print('Done!')
        return temp_files  # type: ignore

    def info_corefile(self):  # type: () -> Optional[list[str]]
        """
        Command to load core dump from file or flash and print it's data in user friendly form
        """
        with self._handle_coredump_loader_error():
            self.exe_elf = ESPCoreDumpElfFile(self.prog)
            core_header_info_dict = self.get_core_header_info_dict(e_machine=self.exe_elf.e_machine)
            self.core_elf = ESPCoreDumpElfFile(core_header_info_dict['core_elf_path'])

        temp_files = core_header_info_dict.pop('temp_files')
        self.chip = self.verify_target(core_header_info_dict)

        if self.exe_elf.e_machine != self.core_elf.e_machine:
            raise ValueError('The arch should be the same between core elf and exe elf')

        task_info, extra_note = self.get_task_info_extra_note_tuple()

        print('===============================================================')
        print('==================== ESP32 CORE DUMP START ====================')

        gdb_args = self.get_gdb_args(is_dbg_mode=False, **core_header_info_dict)

        self.gdb_esp = EspGDB(gdb_args, timeout_sec=self.gdb_timeout_sec)

        extra_info = None
        if extra_note:
            extra_info = Struct('regs' / GreedyRange(Int32ul)).parse(extra_note.desc).regs
            marker = extra_info[0]
            self.print_crashed_task_info(marker)
            self.print_isr_context(extra_info)

        panic_details = self.get_panic_details()
        if panic_details:
            print('Panic reason: ' + panic_details.desc.decode('utf-8'))

        print('\n================== CURRENT THREAD REGISTERS ===================')
        # Only xtensa have exception registers
        self.print_current_thread_registers(extra_note, extra_info)

        print('\n==================== CURRENT THREAD STACK =====================')
        self.print_current_thread_stack(task_info)
        print('\n======================== THREADS INFO =========================')
        self.print_threads_info(task_info)
        print('\n\n======================= ALL MEMORY REGIONS ========================')
        self.print_all_memory_regions()

        if self.print_mem:
            print('\n====================== CORE DUMP MEMORY CONTENTS ========================')
            self.print_core_dump_memory_contents()

        print('\n===================== ESP32 CORE DUMP END =====================')
        print('===============================================================')

        del self.gdb_esp
        print('Done!')
        return temp_files  # type: ignore



#
# SPDX-FileCopyrightText: 2022-2024 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
import os.path

from esp_coredump import CoreDump, __version__
from esp_coredump.cli_ext import parser


def main():
    print(f'espcoredump.py v{__version__}', flush=True)

    args = parser.parse_args()
    debug = getattr(args, 'debug', 3)
    if debug == 0:
        log_level = logging.CRITICAL
    elif debug == 1:
        log_level = logging.ERROR
    elif debug == 2:
        log_level = logging.WARNING
    elif debug == 3:
        log_level = logging.INFO
    else:
        log_level = logging.DEBUG
    logging.basicConfig(format='%(levelname)s: %(message)s', level=log_level)

    kwargs = {k: v for k, v in vars(args).items() if v is not None}

    kwargs.pop('debug', None)
    kwargs.pop('operation', None)

    espcoredump = CoreDump(**kwargs)
    temp_core_files = None

    try:
        if args.operation == 'info_corefile':
            temp_core_files = espcoredump.info_corefile()
        elif args.operation == 'dbg_corefile':
            temp_core_files = espcoredump.dbg_corefile()
    finally:
        if temp_core_files:
            for f in temp_core_files:
                try:
                    os.remove(f)
                except OSError:
                    pass


if __name__ == '__main__':
    main()



SOC_IRAM_LOW = 0x40020000
SOC_IRAM_HIGH = 0x40070000
SOC_DRAM_LOW = 0x3ffb0000
SOC_DRAM_HIGH = 0x40000000
SOC_RTC_DRAM_LOW = 0x3ff9e000
SOC_RTC_DRAM_HIGH = 0x3ffa0000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50002000



SOC_DRAM_LOW = 0x3ffae000
SOC_DRAM_HIGH = 0x40000000
SOC_IRAM_LOW = 0x40080000
SOC_IRAM_HIGH = 0x400a0000
SOC_RTC_DRAM_LOW = 0x3ff80000
SOC_RTC_DRAM_HIGH = 0x3ff82000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50002000



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Optional, Tuple

from construct import Int16ul, Int32ul, Int64ul, Struct

from . import BaseArchMethodsMixin, BaseTargetMethods, ESPCoreDumpLoaderError

INVALID_CAUSE_VALUE = 0xFFFF
XCHAL_EXCCAUSE_NUM = 64

# Exception cause dictionary to get translation of exccause register
# From 4.4.1.5 table 4-64 Exception Causes of Xtensa
# Instruction Set Architecture (ISA) Reference Manual

XTENSA_EXCEPTION_CAUSE_DICT = {
    0: ('IllegalInstructionCause', 'Illegal instruction'),
    1: ('SyscallCause', 'SYSCALL instruction'),
    2: ('InstructionFetchErrorCause',
        'Processor internal physical address or data error during instruction fetch. (See EXCVADDR for more information)'),
    3: ('LoadStoreErrorCause',
        'Processor internal physical address or data error during load or store. (See EXCVADDR for more information)'),
    4: ('Level1InterruptCause', 'Level-1 interrupt as indicated by set level-1 bits in the INTERRUPT register'),
    5: ('AllocaCause', 'MOVSP instruction, if caller`s registers are not in the register file'),
    6: ('IntegerDivideByZeroCause', 'QUOS: QUOU, REMS: or REMU divisor operand is zero'),
    8: ('PrivilegedCause', 'Attempt to execute a privileged operation when CRING ? 0'),
    9: ('LoadStoreAlignmentCause', 'Load or store to an unaligned address. (See EXCVADDR for more information)'),
    12: ('InstrPIFDataErrorCause', 'PIF data error during instruction fetch. (See EXCVADDR for more information)'),
    13: ('LoadStorePIFDataErrorCause',
         'Synchronous PIF data error during LoadStore access. (See EXCVADDR for more information)'),
    14: ('InstrPIFAddrErrorCause', 'PIF address error during instruction fetch. (See EXCVADDR for more information)'),
    15: ('LoadStorePIFAddrErrorCause',
         'Synchronous PIF address error during LoadStore access. (See EXCVADDR for more information)'),
    16: ('InstTLBMissCause', 'Error during Instruction TLB refill. (See EXCVADDR for more information)'),
    17: ('InstTLBMultiHitCause', 'Multiple instruction TLB entries matched. (See EXCVADDR for more information)'),
    18: ('InstFetchPrivilegeCause',
         'An instruction fetch referenced a virtual address at a ring level less than CRING. (See EXCVADDR for more information)'),
    20: ('InstFetchProhibitedCause',
         'An instruction fetch referenced a page mapped with an attribute that does not permit instruction fetch (EXCVADDR).'),
    24: ('LoadStoreTLBMissCause', 'Error during TLB refill for a load or store. (See EXCVADDR for more information)'),
    25: ('LoadStoreTLBMultiHitCause',
         'Multiple TLB entries matched for a load or store. (See EXCVADDR for more information)'),
    26: ('LoadStorePrivilegeCause',
         'A load or store referenced a virtual address at a ring level less than CRING. (See EXCVADDR for more information)'),
    28: ('LoadProhibitedCause',
         'A load referenced a page mapped with an attribute that does not permit loads. (See EXCVADDR for more information)'),
    29: ('StoreProhibitedCause',
         'A store referenced a page mapped with an attribute that does not permit stores [Region Protection Option or MMU Option].'),
    32: ('Coprocessor0Disabled', 'Coprocessor 0 instruction when cp0 disabled'),
    33: ('Coprocessor1Disabled', 'Coprocessor 1 instruction when cp1 disabled'),
    34: ('Coprocessor2Disabled', 'Coprocessor 2 instruction when cp2 disabled'),
    35: ('Coprocessor3Disabled', 'Coprocessor 3 instruction when cp3 disabled'),
    36: ('Coprocessor4Disabled', 'Coprocessor 4 instruction when cp4 disabled'),
    37: ('Coprocessor5Disabled', 'Coprocessor 5 instruction when cp5 disabled'),
    38: ('Coprocessor6Disabled', 'Coprocessor 6 instruction when cp6 disabled'),
    39: ('Coprocessor7Disabled', 'Coprocessor 7 instruction when cp7 disabled'),
    INVALID_CAUSE_VALUE: (
        'InvalidCauseRegister', 'Invalid EXCCAUSE register value or current task is broken and was skipped'),
    # ESP panic pseudo reasons
    XCHAL_EXCCAUSE_NUM + 0: ('UnknownException', 'Unknown exception'),
    XCHAL_EXCCAUSE_NUM + 1: ('DebugException', 'Unhandled debug exception'),
    XCHAL_EXCCAUSE_NUM + 2: ('DoubleException', 'Double exception'),
    XCHAL_EXCCAUSE_NUM + 3: ('KernelException', 'Unhandled kernel exception'),
    XCHAL_EXCCAUSE_NUM + 4: ('CoprocessorException', 'Coprocessor exception'),
    XCHAL_EXCCAUSE_NUM + 5: ('InterruptWDTTimoutCPU0', 'Interrupt wdt timeout on CPU0'),
    XCHAL_EXCCAUSE_NUM + 6: ('InterruptWDTTimoutCPU1', 'Interrupt wdt timeout on CPU1'),
    XCHAL_EXCCAUSE_NUM + 7: ('CacheError', 'Cache disabled but cached memory region accessed'),
}


class ExceptionRegisters(object):
    # extra regs IDs used in EXTRA_INFO note
    EXCCAUSE_IDX = 0
    EXCVADDR_IDX = 1
    EPC1_IDX = 177
    EPC2_IDX = 178
    EPC3_IDX = 179
    EPC4_IDX = 180
    EPC5_IDX = 181
    EPC6_IDX = 182
    EPC7_IDX = 183
    EPS2_IDX = 194
    EPS3_IDX = 195
    EPS4_IDX = 196
    EPS5_IDX = 197
    EPS6_IDX = 198
    EPS7_IDX = 199

    @property
    def registers(self):  # type: () -> dict[str, int]
        return {k: v for k, v in self.__class__.__dict__.items()
                if not k.startswith('__') and isinstance(v, int)}


# Following structs are based on source code
# IDF_PATH/components/espcoredump/src/core_dump_port.c
PrStatus = Struct(
    'si_signo' / Int32ul,
    'si_code' / Int32ul,
    'si_errno' / Int32ul,
    'pr_cursig' / Int16ul,
    'pr_pad0' / Int16ul,
    'pr_sigpend' / Int32ul,
    'pr_sighold' / Int32ul,
    'pr_pid' / Int32ul,
    'pr_ppid' / Int32ul,
    'pr_pgrp' / Int32ul,
    'pr_sid' / Int32ul,
    'pr_utime' / Int64ul,
    'pr_stime' / Int64ul,
    'pr_cutime' / Int64ul,
    'pr_cstime' / Int64ul,
)


def print_exc_regs_info(extra_info):  # type: (list[int]) -> None
    """
    Print the register info by parsing extra_info
    :param extra_info: extra info data str
    :return: None
    """
    exccause = extra_info[1 + 2 * ExceptionRegisters.EXCCAUSE_IDX + 1]
    exccause_str = XTENSA_EXCEPTION_CAUSE_DICT.get(exccause)
    if not exccause_str:
        exccause_str = ('Invalid EXCCAUSE code', 'Invalid EXCAUSE description or not found.')
    print('exccause       0x%x (%s)' % (exccause, exccause_str[0]))
    print('excvaddr       0x%x' % extra_info[1 + 2 * ExceptionRegisters.EXCVADDR_IDX + 1])

    # skip crashed_task_tcb, exccause, and excvaddr
    for i in range(5, len(extra_info), 2):
        if (extra_info[i] >= ExceptionRegisters.EPC1_IDX and extra_info[i] <= ExceptionRegisters.EPC7_IDX):
            print('epc%d           0x%x' % ((extra_info[i] - ExceptionRegisters.EPC1_IDX + 1), extra_info[i + 1]))

    # skip crashed_task_tcb, exccause, and excvaddr
    for i in range(5, len(extra_info), 2):
        if (extra_info[i] >= ExceptionRegisters.EPS2_IDX and extra_info[i] <= ExceptionRegisters.EPS7_IDX):
            print('eps%d           0x%x' % ((extra_info[i] - ExceptionRegisters.EPS2_IDX + 2), extra_info[i + 1]))


# from "gdb/xtensa-tdep.h"
# typedef struct
# {
# 0    xtensa_elf_greg_t pc;
# 1    xtensa_elf_greg_t ps;
# 2    xtensa_elf_greg_t lbeg;
# 3    xtensa_elf_greg_t lend;
# 4    xtensa_elf_greg_t lcount;
# 5    xtensa_elf_greg_t sar;
# 6    xtensa_elf_greg_t windowstart;
# 7    xtensa_elf_greg_t windowbase;
# 8..63 xtensa_elf_greg_t reserved[8+48];
# 64   xtensa_elf_greg_t ar[64];
# } xtensa_elf_gregset_t;
REG_PC_IDX = 0
REG_PS_IDX = 1
REG_LB_IDX = 2
REG_LE_IDX = 3
REG_LC_IDX = 4
REG_SAR_IDX = 5
# REG_WS_IDX = 6
# REG_WB_IDX = 7
REG_AR_START_IDX = 64
# REG_AR_NUM = 64
# FIXME: acc to xtensa_elf_gregset_t number of regs must be 128,
# but gdb complains when it less then 129
REG_NUM = 129

# XT_SOL_EXIT = 0
XT_SOL_PC = 1
XT_SOL_PS = 2
# XT_SOL_NEXT = 3
XT_SOL_AR_START = 4
XT_SOL_AR_NUM = 4
# XT_SOL_FRMSZ = 8

XT_STK_EXIT = 0
XT_STK_PC = 1
XT_STK_PS = 2
XT_STK_AR_START = 3
XT_STK_AR_NUM = 16
XT_STK_SAR = 19
XT_STK_EXCCAUSE = 20
XT_STK_EXCVADDR = 21
XT_STK_LBEG = 22
XT_STK_LEND = 23
XT_STK_LCOUNT = 24
XT_STK_FRMSZ = 25


class XtensaMethodsMixin(BaseArchMethodsMixin):
    @staticmethod
    def get_registers_from_stack(data, grows_down):
        # type: (bytes, bool) -> Tuple[list[int], Optional[dict[int, int]]]
        extra_regs = {v: 0 for v in ExceptionRegisters().registers.values()}
        regs = [0] * REG_NUM
        # TODO: support for growing up stacks
        if not grows_down:
            raise ESPCoreDumpLoaderError('Growing up stacks are not supported for now!')
        ex_struct = Struct(
            'stack' / Int32ul[XT_STK_FRMSZ]
        )
        if len(data) < ex_struct.sizeof():
            raise ESPCoreDumpLoaderError('Too small stack to keep frame: %d bytes!' % len(data))

        stack = ex_struct.parse(data).stack
        # Stack frame type indicator is always the first item
        rc = stack[XT_STK_EXIT]
        if rc != 0:
            regs[REG_PC_IDX] = stack[XT_STK_PC]
            regs[REG_PS_IDX] = stack[XT_STK_PS]
            for i in range(XT_STK_AR_NUM):
                regs[REG_AR_START_IDX + i] = stack[XT_STK_AR_START + i]
            regs[REG_SAR_IDX] = stack[XT_STK_SAR]
            regs[REG_LB_IDX] = stack[XT_STK_LBEG]
            regs[REG_LE_IDX] = stack[XT_STK_LEND]
            regs[REG_LC_IDX] = stack[XT_STK_LCOUNT]
            # FIXME: crashed and some running tasks (e.g. prvIdleTask) have EXCM bit set
            #   and GDB can not unwind callstack properly (it implies not windowed call0)
            if regs[REG_PS_IDX] & (1 << 5):
                regs[REG_PS_IDX] &= ~(1 << 4)
            if stack[XT_STK_EXCCAUSE] in XTENSA_EXCEPTION_CAUSE_DICT:
                extra_regs[ExceptionRegisters.EXCCAUSE_IDX] = stack[XT_STK_EXCCAUSE]
            else:
                extra_regs[ExceptionRegisters.EXCCAUSE_IDX] = INVALID_CAUSE_VALUE
            extra_regs[ExceptionRegisters.EXCVADDR_IDX] = stack[XT_STK_EXCVADDR]
        else:
            regs[REG_PC_IDX] = stack[XT_SOL_PC]
            regs[REG_PS_IDX] = stack[XT_SOL_PS]
            for i in range(XT_SOL_AR_NUM):
                regs[REG_AR_START_IDX + i] = stack[XT_SOL_AR_START + i]
            # nxt = stack[XT_SOL_NEXT]
        return regs, extra_regs

    @staticmethod
    def build_prstatus_data(tcb_addr, task_regs):  # type: (int, list[int]) -> Any
        return PrStatus.build({
            'si_signo': 0,
            'si_code': 0,
            'si_errno': 0,
            'pr_cursig': 0,  # TODO: set sig only for current/failed task
            'pr_pad0': 0,
            'pr_sigpend': 0,
            'pr_sighold': 0,
            'pr_pid': tcb_addr,
            'pr_ppid': 0,
            'pr_pgrp': 0,
            'pr_sid': 0,
            'pr_utime': 0,
            'pr_stime': 0,
            'pr_cutime': 0,
            'pr_cstime': 0,
        }) + Int32ul[len(task_regs)].build(task_regs)


class Esp32Methods(BaseTargetMethods, XtensaMethodsMixin):
    TARGET = 'esp32'


class Esp32S2Methods(BaseTargetMethods, XtensaMethodsMixin):
    TARGET = 'esp32s2'


class Esp32S3Methods(BaseTargetMethods, XtensaMethodsMixin):
    TARGET = 'esp32s3'



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
import os
import re
import time
from subprocess import TimeoutExpired

from pygdbmi.gdbcontroller import GdbController

from . import ESPCoreDumpError

DEFAULT_GDB_TIMEOUT_SEC = 3


class EspGDB(object):
    def __init__(self, gdb_args, timeout_sec=DEFAULT_GDB_TIMEOUT_SEC):

        """
        Start GDB and initialize a GdbController instance
        """
        try:
            self.p = GdbController(command=gdb_args)
        except TypeError:
            # fallback for pygdbmi<0.10.0.0.
            self.p = GdbController(gdb_path=gdb_args[0], gdb_args=gdb_args[1:])

        self.timeout = timeout_sec

        # Consume initial output by issuing a dummy command
        self._gdbmi_run_cmd_get_responses(cmd='-data-list-register-values x pc',
                                          resp_message=None, resp_type='console', multiple=True,
                                          done_message='done', done_type='result')

    def __del__(self):
        """EspGDB destructor taking GdbController.gdb_process.exit() and adjusting it to work properly"""
        try:
            if self.p.gdb_process:
                self.p.gdb_process.terminate()
                if os.name != 'nt':
                    # this is causing an Exception on Windows, but is required on UNIX systems
                    self.p.gdb_process.communicate()  # Close pipes (STDIN, STDOUT, STDERR)
                try:
                    self.p.gdb_process.wait(timeout=1)
                except TimeoutExpired:
                    if self.p.gdb_process.returncode is None:
                        self.p.gdb_process.kill()
                self.p.gdb_process = None
        except IndexError:
            logging.warning('Attempt to terminate the GDB process failed, because it is already terminated. Skip')

    def _gdbmi_run_cmd_get_responses(self, cmd, resp_message, resp_type, multiple=True,
                                     done_message=None, done_type=None, response_delay_sec=None):

        self.p.write(cmd, read_response=False)
        t_end = time.time() + (response_delay_sec or self.timeout)
        filtered_response_list = []
        all_responses = []
        while time.time() < t_end:
            more_responses = self.p.get_gdb_response(timeout_sec=0, raise_error_on_timeout=False)
            filtered_response_list += filter(lambda rsp: rsp['message'] == resp_message and rsp['type'] == resp_type,
                                             more_responses)
            all_responses += more_responses
            if filtered_response_list and not multiple:
                break
            if done_message and done_type and self._gdbmi_filter_responses(more_responses, done_message, done_type):
                break
        if not filtered_response_list and not multiple:
            raise ESPCoreDumpError("Couldn't find response with message '{}', type '{}' in responses '{}'".format(
                resp_message, resp_type, str(all_responses)
            ))
        return filtered_response_list

    def _gdbmi_run_cmd_get_one_response(self, cmd, resp_message, resp_type, response_delay_sec=None):

        return self._gdbmi_run_cmd_get_responses(cmd, resp_message, resp_type, response_delay_sec=response_delay_sec,
                                                 multiple=False)[0]

    def _gdbmi_data_evaluate_expression(self, expr):
        """ Get the value of an expression, similar to the 'print' command """
        return self._gdbmi_run_cmd_get_one_response("-data-evaluate-expression \"%s\"" % expr,
                                                    'done', 'result')['payload']['value']

    def get_tcb_variable(self, tcb_addr, variable):
        """ Get FreeRTOS variable from given TCB address """
        try:
            val = self._gdbmi_data_evaluate_expression('(char*)((TCB_t *)0x%x)->%s' % (tcb_addr, variable))
        except (ESPCoreDumpError, KeyError):
            # KeyError is raised when "value" is not in "payload"
            return ''
        return val

    def parse_tcb_variable(self, tcb_addr, variable):
        """ Get FreeRTOS variable from given TCB address """
        val = self.get_tcb_variable(tcb_addr, variable)

        # Value is of form '0x12345678 ""'
        result = re.search(r'0x[0-9a-fA-F]+', val)
        if result:
            return result.group(0)
        return ''

    def get_freertos_task_name(self, tcb_addr):
        """ Get FreeRTOS task name given the TCB address """
        val = self.get_tcb_variable(tcb_addr, 'pcTaskName')

        # Value is of form '0x12345678 "task_name"', extract the actual name
        result = re.search(r"\"([^']*)\"$", val)
        if result:
            return result.group(1)
        return ''

    def run_cmd(self, gdb_cmd):
        """ Execute a generic GDB console command via MI2
        """
        filtered_responses = self._gdbmi_run_cmd_get_responses(cmd="-interpreter-exec console \"%s\"" % gdb_cmd,
                                                               resp_message=None, resp_type='console', multiple=True,
                                                               done_message='done', done_type='result')
        return ''.join([x['payload'] for x in filtered_responses]) \
            .replace('\\n', '\n') \
            .replace('\\t', '\t') \
            .rstrip('\n') \
            .replace('\\"', '"')

    def get_thread_info(self, response_delay_sec=DEFAULT_GDB_TIMEOUT_SEC):
        """ Get information about all threads known to GDB, and the current thread ID """
        result = self._gdbmi_run_cmd_get_one_response('-thread-info', 'done', 'result', response_delay_sec=response_delay_sec)['payload']
        if not result:
            return None, None

        current_thread_id = result['current-thread-id']
        threads = result['threads']
        return threads, current_thread_id

    def switch_thread(self, thr_id):
        """ Tell GDB to switch to a specific thread, given its ID """
        self._gdbmi_run_cmd_get_one_response('-thread-select %s' % thr_id, 'done', 'result')

    @staticmethod
    def _gdbmi_filter_responses(responses, resp_message, resp_type):
        return list(filter(lambda rsp: rsp['message'] == resp_message and rsp['type'] == resp_type, responses))

    @staticmethod
    def gdb2freertos_thread_id(gdb_target_id):
        """ Convert GDB 'target ID' to the FreeRTOS TCB address """
        return int(gdb_target_id.replace('process ', ''), 0)



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import os
import sys

try:
    import esptool  # noqa: F401
except ImportError:
    idf_path = os.getenv('IDF_PATH')
    if not idf_path or not os.path.exists(idf_path):
        raise ModuleNotFoundError('Please read the README file for the latest installation instructions')
    sys.path.insert(0, os.path.join(idf_path, 'components', 'esptool_py', 'esptool'))
    import esptool  # noqa: F401

from .coredump import CoreDump

__all__ = [
    'CoreDump',
]

__version__ = '1.13.1'



SOC_IRAM_LOW = 0x40370000
SOC_IRAM_HIGH = 0x403e0000
SOC_DRAM_LOW = 0x3fc88000
SOC_DRAM_HIGH = 0x3fd00000
SOC_RTC_DRAM_LOW = 0x600fe000
SOC_RTC_DRAM_HIGH = 0x60100000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50002000



SOC_IRAM_LOW = 0x40800000
SOC_IRAM_HIGH = 0x40880000
SOC_DRAM_LOW = 0x40800000
SOC_DRAM_HIGH = 0x40880000
SOC_RTC_DRAM_LOW = 0x50000000
SOC_RTC_DRAM_HIGH = 0x50004000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50004000



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#
# This script will parse soc.h and generate .py file containing all of the target constants.
# Once one manually adds a new target, the constants file can be regenerated.
# The script is ESP-IDF dependent and is not intended for use by end users.
#
import os
import sys
from ast import literal_eval

from esp_coredump.coredump import IDF_SETUP_ERROR
from esp_coredump.corefile import SUPPORTED_TARGETS

IDF_PATH = os.getenv('IDF_PATH')


def main():  # type: () -> None
    constants = [
        'SOC_DRAM_LOW',
        'SOC_DRAM_HIGH',
        'SOC_IRAM_LOW',
        'SOC_IRAM_HIGH',
        'SOC_RTC_DATA_LOW',
        'SOC_RTC_DATA_HIGH',
        'SOC_RTC_DRAM_LOW',
        'SOC_RTC_DRAM_HIGH',
    ]

    for target in SUPPORTED_TARGETS:
        target_constants = {}
        soc_header_fp = os.path.join(IDF_PATH, 'components/soc/{}/include/soc/soc.h'.format(target))  # type: ignore
        module_fp = os.path.join(os.path.normpath(os.path.abspath(os.path.dirname(__file__))),
                                 'soc_headers',
                                 '{}.py'.format(target)
                                 )

        with open(soc_header_fp) as fr:
            for line in fr.readlines():
                for attr in constants:
                    if '#define {}'.format(attr) in line:
                        target_constants[attr] = literal_eval(line.strip().split()[-1])

        for attr in constants:
            if attr not in target_constants:
                print(f'WARNING: {attr} is missing in {soc_header_fp}, check if this is correct!', file=sys.stderr)
                target_constants[attr] = 2**32 - 1

        with open(module_fp, 'w') as fw:
            for k, v in target_constants.items():
                fw.write('{} = {}\n'.format(k, hex(v).lower()))


if __name__ == '__main__':
    if not IDF_PATH:
        print(IDF_SETUP_ERROR)
        sys.exit(1)
    main()



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#

import hashlib
import os
from typing import Optional

from construct import (AlignedStruct, Bytes, Const, Container, GreedyRange,
                       Int16ul, Int32ul, Padding, Pointer, Sequence, Struct,
                       this)

# Following structs are based on spec
# https://refspecs.linuxfoundation.org/elf/elf.pdf
# and source code
# IDF_PATH/components/espcoredump/include_core_dump/elf.h

ElfIdentification = Struct(
    'EI_MAG' / Const(b'\x7fELF'),
    'EI_CLASS' / Const(b'\x01'),  # ELFCLASS32
    'EI_DATA' / Const(b'\x01'),  # ELFDATA2LSB
    'EI_VERSION' / Const(b'\x01'),  # EV_CURRENT
    Padding(9),
)

ElfHeader = Struct(
    'e_ident' / ElfIdentification,
    'e_type' / Int16ul,
    'e_machine' / Int16ul,
    'e_version' / Int32ul,
    'e_entry' / Int32ul,
    'e_phoff' / Int32ul,
    'e_shoff' / Int32ul,
    'e_flags' / Int32ul,
    'e_ehsize' / Int16ul,
    'e_phentsize' / Int16ul,
    'e_phnum' / Int16ul,
    'e_shentsize' / Int16ul,
    'e_shnum' / Int16ul,
    'e_shstrndx' / Int16ul,
)

SectionHeader = Struct(
    'sh_name' / Int32ul,
    'sh_type' / Int32ul,
    'sh_flags' / Int32ul,
    'sh_addr' / Int32ul,
    'sh_offset' / Int32ul,
    'sh_size' / Int32ul,
    'sh_link' / Int32ul,
    'sh_info' / Int32ul,
    'sh_addralign' / Int32ul,
    'sh_entsize' / Int32ul,
)

ProgramHeader = Struct(
    'p_type' / Int32ul,
    'p_offset' / Int32ul,
    'p_vaddr' / Int32ul,
    'p_paddr' / Int32ul,
    'p_filesz' / Int32ul,
    'p_memsz' / Int32ul,
    'p_flags' / Int32ul,
    'p_align' / Int32ul,
)

ElfHeaderTables = Struct(
    'elf_header' / ElfHeader,
    'program_headers' / Pointer(this.elf_header.e_phoff, ProgramHeader[this.elf_header.e_phnum]),
    'section_headers' / Pointer(this.elf_header.e_shoff, SectionHeader[this.elf_header.e_shnum]),
)

NoteSection = AlignedStruct(
    4,
    'namesz' / Int32ul,
    'descsz' / Int32ul,
    'type' / Int32ul,
    'name' / Bytes(this.namesz),
    'desc' / Bytes(this.descsz),
)

NoteSections = GreedyRange(NoteSection)


class ElfFile(object):
    """
    Elf class to a single elf file
    """

    SHN_UNDEF = 0x00
    SHT_PROGBITS = 0x01
    SHT_STRTAB = 0x03
    SHT_NOBITS = 0x08

    PT_LOAD = 0x01
    PT_NOTE = 0x04

    ET_CORE = 0x04

    EV_CURRENT = 0x01

    def __init__(self, elf_path=None, e_type=None, e_machine=None):
        # type: (Optional[str], Optional[int], Optional[int]) -> None
        self.e_type = e_type
        self.e_machine = e_machine

        self._struct = None  # type: Optional[Struct]
        self._model = None  # type: Optional[Container]

        self.sections = []  # type: list[ElfSection]
        self.load_segments = []  # type: list[ElfSegment]
        self.note_segments = []  # type: list[ElfNoteSegment]

        self.sha256 = b''  # type: bytes

        if elf_path and os.path.isfile(elf_path):
            self.read_elf(elf_path)

    def read_elf(self, elf_path):  # type: (str) -> None
        """
        Read elf file, also write to ``self.model``, ``self.program_headers``,
        ``self.section_headers``
        :param elf_path: elf file path
        :return: None
        """
        with open(elf_path, 'rb') as fr:
            elf_bytes = fr.read()
        header_tables = ElfHeaderTables.parse(elf_bytes)
        self.e_type = header_tables.elf_header.e_type
        self.e_machine = header_tables.elf_header.e_machine

        self._struct = self._generate_struct_from_headers(header_tables)
        self._model = self._struct.parse(elf_bytes)

        self.load_segments = [ElfSegment(seg.ph.p_vaddr,
                                         seg.data,
                                         seg.ph.p_flags) for seg in self._model.load_segments]
        self.note_segments = [ElfNoteSegment(seg.ph.p_vaddr,
                                             seg.data,
                                             seg.ph.p_flags) for seg in self._model.note_segments]
        self.sections = [ElfSection(self._parse_string_table(self._model.string_table, sec.sh.sh_name),
                                    sec.sh.sh_addr,
                                    sec.data,
                                    sec.sh.sh_flags) for sec in self._model.sections]

        # calculate sha256 of the input bytes (note: this may not be the same as the sha256 of any generated
        # output struct, as the ELF parser may change some details.)
        sha256 = hashlib.sha256()
        sha256.update(elf_bytes)
        self.sha256 = sha256.digest()

    @staticmethod
    def _parse_string_table(byte_str, offset):  # type: (bytes, int) -> str
        section_name_str = byte_str[offset:]
        string_end = section_name_str.find(0x00)

        if (string_end == -1):
            raise ValueError('Unable to get section name from section header string table')

        name = section_name_str[:string_end].decode('utf-8')

        return name

    def _generate_struct_from_headers(self, header_tables):  # type: (Container) -> Struct
        """
        Generate ``construct`` Struct for this file
        :param header_tables: contains elf_header, program_headers, section_headers
        :return: Struct of the whole file
        """
        elf_header = header_tables.elf_header
        program_headers = header_tables.program_headers
        section_headers = header_tables.section_headers
        assert program_headers or section_headers

        string_table_sh = None
        load_segment_subcons = []
        note_segment_subcons = []
        # Here we point back to make segments know their program headers
        for i, ph in enumerate(program_headers):
            args = [
                'ph' / Pointer(elf_header.e_phoff + i * ProgramHeader.sizeof(), ProgramHeader),
                'data' / Pointer(ph.p_offset, Bytes(ph.p_filesz)),
            ]
            if ph.p_vaddr == 0 and ph.p_type == self.PT_NOTE:
                args.append('note_secs' / Pointer(ph.p_offset, NoteSections))
                note_segment_subcons.append(Struct(*args))
            elif ph.p_vaddr != 0:
                load_segment_subcons.append(Struct(*args))

        section_subcons = []
        for i, sh in enumerate(section_headers):
            if sh.sh_type == self.SHT_STRTAB and i == elf_header.e_shstrndx:
                string_table_sh = sh
            elif sh.sh_addr != 0 and sh.sh_type == self.SHT_PROGBITS:
                section_subcons.append(Struct(
                    'sh' / Pointer(elf_header.e_shoff + i * SectionHeader.sizeof(), SectionHeader),
                    'data' / Pointer(sh.sh_offset, Bytes(sh.sh_size)),
                ))

        args = [
            'elf_header' / ElfHeader,
            'load_segments' / Sequence(*load_segment_subcons),
            'note_segments' / Sequence(*note_segment_subcons),
            'sections' / Sequence(*section_subcons),
        ]
        if string_table_sh is not None:
            args.append('string_table' / Pointer(string_table_sh.sh_offset, Bytes(string_table_sh.sh_size)))

        return Struct(*args)


class ElfSection(object):
    SHF_WRITE = 0x01
    SHF_ALLOC = 0x02
    SHF_EXECINSTR = 0x04
    SHF_MASKPROC = 0xf0000000

    def __init__(self, name, addr, data, flags):  # type: (str, int, bytes, int) -> None
        self.name = name
        self.addr = addr
        self.data = data
        self.flags = flags

    def attr_str(self):  # type: () -> str
        if self.flags & self.SHF_MASKPROC:
            return 'MS'

        res = 'R'
        res += 'W' if self.flags & self.SHF_WRITE else ' '
        res += 'X' if self.flags & self.SHF_EXECINSTR else ' '
        res += 'A' if self.flags & self.SHF_ALLOC else ' '
        return res

    def __repr__(self):  # type: () -> str
        return '{:>32} [Addr] 0x{:>08X}, [Size] 0x{:>08X} {:>4}' \
            .format(self.name, self.addr, len(self.data), self.attr_str())


class ElfSegment(object):
    PF_X = 0x01
    PF_W = 0x02
    PF_R = 0x04

    def __init__(self, addr, data, flags):  # type: (int, bytes, int) -> None
        self.addr = addr
        self.data = data
        self.flags = flags
        self.type = ElfFile.PT_LOAD

    def attr_str(self):  # type: () -> str
        res = ''
        res += 'R' if self.flags & self.PF_R else ' '
        res += 'W' if self.flags & self.PF_W else ' '
        res += 'E' if self.flags & self.PF_X else ' '
        return res

    @staticmethod
    def _type_str():  # type: () -> str
        return 'LOAD'

    def __repr__(self):  # type: () -> str
        return '{:>8} Addr 0x{:>08X}, Size 0x{:>08X} Flags {:4}' \
            .format(self._type_str(), self.addr, len(self.data), self.attr_str())


class ElfNoteSegment(ElfSegment):
    def __init__(self, addr, data, flags):  # type: (int, bytes, int) -> None
        super(ElfNoteSegment, self).__init__(addr, data, flags)
        self.type = ElfFile.PT_NOTE
        self.note_secs = NoteSections.parse(self.data)
        for note in self.note_secs:
            # note.name should include a terminating NUL byte, plus possible
            # padding
            #
            # (note: construct.PaddingString can't parse this if there
            # are non-zero padding bytes after the NUL, it also parses those.)
            note.name = note.name.split(b'\x00')[0]

    @staticmethod
    def _type_str():  # type: () -> str
        return 'NOTE'


TASK_STATUS_CORRECT = 0x00
TASK_STATUS_TCB_CORRUPTED = 0x01
TASK_STATUS_STACK_CORRUPTED = 0x02

EspTaskStatus = Struct(
    'task_index' / Int32ul,
    'task_flags' / Int32ul,
    'task_tcb_addr' / Int32ul,
    'task_stack_start' / Int32ul,
    'task_stack_len' / Int32ul,
    'task_name' / Bytes(16),
)


class ESPCoreDumpElfFile(ElfFile):
    PT_ESP_INFO = 8266
    PT_ESP_PANIC_DETAILS = 679
    PT_ESP_TASK_INFO = 678
    PT_ESP_EXTRA_INFO = 677

    CURR_TASK_MARKER = 0xdeadbeef

    # ELF file machine type
    EM_XTENSA = 0x5E
    EM_RISCV = 0xF3

    def __init__(self, elf_path=None, e_type=None, e_machine=None):
        # type: (Optional[str], Optional[int], Optional[int]) -> None
        _e_type = e_type or self.ET_CORE
        _e_machine = e_machine or self.EM_XTENSA
        super(ESPCoreDumpElfFile, self).__init__(elf_path, _e_type, _e_machine)

    def add_segment(self, addr, data, seg_type, flags):  # type: (int, bytes, int, int) -> None
        if seg_type != self.PT_NOTE:
            self.load_segments.append(ElfSegment(addr, data, flags))
        else:
            self.note_segments.append(ElfNoteSegment(addr, data, flags))

    def dump(self, output_path):  # type: (str) -> None
        """
        Dump self.model into file
        :param output_path: output file path
        :return: None
        """
        res = b''
        res += ElfHeader.build({
            'e_type': self.e_type,
            'e_machine': self.e_machine,
            'e_version': self.EV_CURRENT,
            'e_entry': 0,
            'e_phoff': ElfHeader.sizeof(),
            'e_shoff': 0,
            'e_flags': 0,
            'e_ehsize': ElfHeader.sizeof(),
            'e_phentsize': ProgramHeader.sizeof(),
            'e_phnum': len(self.load_segments) + len(self.note_segments),
            'e_shentsize': 0,
            'e_shnum': 0,
            'e_shstrndx': self.SHN_UNDEF,
        })

        offset = ElfHeader.sizeof() + (len(self.load_segments) + len(self.note_segments)) * ProgramHeader.sizeof()
        _segments = self.load_segments + self.note_segments  # type: ignore
        for seg in _segments:
            res += ProgramHeader.build({
                'p_type': seg.type,
                'p_offset': offset,
                'p_vaddr': seg.addr,
                'p_paddr': seg.addr,
                'p_filesz': len(seg.data),
                'p_memsz': len(seg.data),
                'p_flags': seg.flags,
                'p_align': 0,
            })
            offset += len(seg.data)

        for seg in _segments:
            res += seg.data

        with open(output_path, 'wb') as fw:
            fw.write(res)



#
# SPDX-FileCopyrightText: 2022-2023 Espressif Systems (Shanghai) CO LTD
#
# SPDX-License-Identifier: Apache-2.0
#
import json


class FatalError(Exception):
    pass


def load_json_from_file(filename: str) -> dict:
    with open(filename, 'r') as file:
        data: dict = json.load(file)
    return data
